import numpy
import numpy as np
from numpy import *
'''带扩展功能，比如fftshift    fftabs    z中心以及相位'''

def fft_traces(self,traces):
    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if traces.ndim == 1:
        N = len(traces)
        return np.array(1.0 / N * np.fft.fftshift(np.fft.fft(traces)))
    if traces.ndim == 2:
        fft_list = []
        for i in range(traces.shape[0]):
            N = traces.shape[1]
            fft_list.append((np.fft.fft(traces[i])))
        return np.array(fft_list)
def ifft_traces(self,traces):

    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if traces.ndim == 1:
        N = len(traces)
        return np.array(1.0 / N * np.fft.ifftshift(np.fft.ifft(traces)))
    if traces.ndim == 2:
        ifft_list = []
        for i in range(traces.shape[0]):
            ifft_list.append((np.fft.ifft(traces[i])))
        return np.array(ifft_list)

if __name__ == '__main__':

    # f = np.fft.fft2(n,axes= - 1)
    f = np.fft.fft2(n)
    #
    # # fshift = np.fft.fftshift(f)
    # # freq = np.fft.fftfreq(n.size, d=1)
    # # res = np.log(np.abs(fshift))
    #
    #
    # plt.plot(f[0])
    # plt.show()
        # pass

    # print(f.ndim)
    # a = np.mgrid[:5, :5][0]

    # b = np.array(range(8))
    # m = fft_oper()
    # import matplotlib.pyplot as plt
    # cc = np.fft.fft(n)
    # # nn = m.ifft_traces(cc)
    # # print(n.shape)
    # # print(m.fft_traces(n))
    # # print(fft_oper.fft_traces(n))
    # print(cc)
    # plt.plot(cc.T)
    # plt.show()


import numpy as np
from matplotlib import pyplot as plt
from tqdm import tqdm, trange

def prepare_data(url_trace, trace_name, url_data, data_name, n,loadData):
    m = {}
    v = {}
    trace = np.load(url_trace + data_name+ r"0.npy")
    count_temp = np.zeros(300, dtype=np.int32)
    for i in loadData:
        m[i] = np.zeros(trace.shape[1])
        v[i] = np.zeros(trace.shape[1])
    for j in trange(n):
        data = np.loadtxt(url_data + data_name + r"{0}.txt".format(j), delimiter=',', dtype="int")
        # print(data)
        trace = np.load(url_trace +trace_name+ r"{0}.npy".format(j))
        for count, label in enumerate(data):
            # d[label].append(trace[count])
            old_mean = m[label]

            m[label] = m[label] + (trace[count] - m[label]) / (count_temp[label] + 1)
            v[label] = v[label] + ((trace[count] - old_mean) * (trace[count] - m[label]) - v[label]) / (
                    count_temp[label] + 1)
            count_temp[label] += 1
    signal = []
    noise = []
    for i_ in loadData:
        signal.append(m[i_])
    for j_ in loadData:
        noise.append(v[j_])
    signal_var = np.var(signal, axis=0)
    noise_mean = np.mean(noise, axis=0)
    return signal_var / noise_mean

def snr_function(n,url_trace,trace_name,url_data,data_name):
    plt.rcParams['figure.figsize'] = (12.0, 4.0)
    f, ax = plt.subplots(1, 1)
    ax.set_title('snr_traces')
    list = (np.unique(np.loadtxt(url_data + data_name+r"0.txt", delimiter=',', dtype="int"))).tolist()
    for i in range(n - 1):
        dataaaa = np.loadtxt(url_data + data_name + r"{0}.txt".format(i + 1), delimiter=',', dtype="int")
        num = (np.unique(dataaaa)).tolist()
        list = list + num
    loadData = (np.unique(list))
    result_data = prepare_data(url_trace, trace_name, url_data, data_name, n,loadData)
    ax.plot(result_data)
    plt.show()


if __name__ == '__main__':
    '''
    确定文件路径，名字，块数    
    '''
    url_trace = r"H:/zhoujiayun4/"
    url_data = r"H:/zhoujiayun4/"
    data_name = "aaadata"
    trace_name = "arrPart"
    n = 100




    # 运行此函数
    snr_function(n,url_trace,trace_name,url_data,data_name)




# 给我一个中间值，就是函数处理的中间值，我去找能量映射模型？
# 汉明重量
import numpy as np
# 有时间加一个类型限制   整数类型哦
def _hamming_weight(intermediate_value):
    '''

    :param intermediate_value: 一个中间值
    :return: 中间值的汉明重量
    '''
    m = 0
    while intermediate_value != 0:
        # print(type(intermediate_value))
        m=m+(intermediate_value & 1)
        intermediate_value >>= 1
    return m
def _LSB(intermediate_value):
    '''

    :param intermediate_value: 一个中间值
    :return: 返回最低有效位
    '''
    m = 0
    while intermediate_value != 0:
        m = intermediate_value & 1
    return m
def HW(data):
    '''

    :param data: 一组中间值
    :return: 一组中间值的汉明重量
    '''
    if not isinstance(data, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if data.ndim == 1:
        new_data = np.zeros(len(data), dtype=float)
        for i in range(len(data)):
            new_data[i] = _hamming_weight(int(data[i]))
        return new_data
    if data.ndim == 2:
        new_data = np.zeros((data.shape(0),data.shape[1]), dtype=float)
        for i in range(data.shape[0]):
            for j in range(data.shape[1]):
                new_data[i][j] = _hamming_weight(int(data[i][j]))
        return new_data
def LSB(data):
    """

    :param data:同上
    :return:
    """
    if not isinstance(data, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if data.ndim == 1:
        new_data = np.zeros(len(data), dtype=float)
        for i in range(len(data)):
            new_data[i] = _LSB(data[i])
        return new_data
    if data.ndim == 2:
        new_data = np.zeros((data.shape(0),data.shape[1]), dtype=float)
        for i in range(data.shape[0]):
            for j in range(data.shape[1]):
                new_data[i][j] = _LSB(data[i][j])
        return new_data
# print(HW(np.array(255)))
if __name__ == '__main__':
    pass


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
'''
@Project ：sca 
@File    ：zhoujiayun.py
@Author  ：suyang
@Date    ：2022/5/21 12:57 
'''
import matplotlib.pyplot as plt
import numpy as np
import time
from tqdm import tqdm, trange

# np.nanmax:排除nan值求最大
np.seterr(divide='ignore', invalid='ignore')
from numba import njit, jit

def getHW(byte):
    '''

    :param byte: input intermediate value
    :return: hw
    '''
    HW_table = [0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4, 1, 2, 2, 3, 2, 3, 3,
                4, 2, 3, 3, 4, 3, 4, 4, 5, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4,
                4, 5, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 1, 2, 2, 3, 2,
                3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5,
                4, 5, 5, 6, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 3, 4, 4,
                5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 1, 2, 2, 3, 2, 3, 3, 4, 2, 3,
                3, 4, 3, 4, 4, 5, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 2,
                3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 3, 4, 4, 5, 4, 5, 5, 6,
                4, 5, 5, 6, 5, 6, 6, 7, 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5,
                6, 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 3, 4, 4, 5, 4, 5,
                5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6,
                7, 7, 8]
    return HW_table[byte]
# all the attack need data matrix
def big_correlation_func(n, url_trace, trace_name,url_data,data_name):
    '''

    :param n: 变量的块数
    :param url_trace: 曲线集的路径
    :param url_data: 数据的路径
    :return:
    '''
    for i in trange(n):
        dataaaa = np.loadtxt(url_data + data_name + r"{0}.npy".format(i))
        for j in range(len(dataaaa)):
            dataaaa[j] = getHW(dataaaa[j])
        np.save(url_data + r"data_hw{0}.npy".format(i), dataaaa)
    arr = np.load(url_trace + trace_name+r"0.npy")
    # data = np.load(url_data + r"new_arrdata0.npy")
    Na = arr.shape[1]
    old_cov = np.zeros(Na)
    old_mean_data = 0
    old_mean_traces = np.zeros(Na)
    old_var_data = 0
    old_var_traces = np.zeros(Na)
    temp = 0
    for j in trange(n):
        data = np.load(url_data + r"data_hw{0}.npy".format(j))
        arr = np.load(url_trace + trace_name + r"{0}.npy".format(j))
        for i in range(arr.shape[0]):
            new_mean_data = old_mean_data + (data[i] - old_mean_data) / (temp + 1)
            new_mean_traces = old_mean_traces + (arr[i] - old_mean_traces) / (temp + 1)

            new_var_data = old_var_data + ((data[i] - old_mean_data) * (data[i] - new_mean_data) - old_var_data) / (
                    temp + 1)
            new_var_traces = old_var_traces + (
                    (arr[i] - old_mean_traces) * (arr[i] - new_mean_traces) - old_var_traces) / (temp + 1)
            new_cov = (old_cov * temp + (data[i] - old_mean_data) * (arr[i] - new_mean_traces)) / (temp + 1)
            old_mean_data = new_mean_data
            old_mean_traces = new_mean_traces
            old_cov = new_cov
            old_var_traces = new_var_traces
            old_var_data = new_var_data
            temp = temp + 1
    correlation_result = old_cov / np.sqrt(old_var_traces * old_var_data)
    return correlation_result

if __name__ == '__main__':
    # 文件块数
    n = 100
    # 所需数据所在文件夹的路径
    url_trace = r"H:/zhoujiayun1/"
    url_data = r"H:/zhoujiayun1/"
    # 所需数据的名字
    data_name = "data"
    trace_name = "data"


    #运行函数且绘图
    f, ax = plt.subplots(1, 1)
    ax.set_title('hw_cpa_traces')
    result = big_correlation_func(n, url_trace, trace_name,url_data,data_name)
    ax.plot(result)
    plt.show()




import numpy as np

from  matplotlib import pyplot as plt
from attackmeasure.new_math import mean_func,var_func
np.seterr(divide='ignore',invalid='ignore')


def ttest(t1,t2):
    Na = t1.shape[0]
    Nb = t2.shape[0]
    temp1 = mean_func(t1)-mean_func(t2)
    temp2 = (var_func(t1) / Na) + (var_func(t2) / Nb)
    test_result = temp1/np.sqrt(temp2)
    return test_result
# def T_test(traces,)

def split_func(traces):
    if traces.ndim == 2:
        n = int(traces.shape[0] / 2)
        print(n)
        traceodd = np.zeros((n, traces.shape[1]))
        traceven = np.zeros((n, traces.shape[1]))
        odd = 0
        even = 0
        for i in range(traces.shape[0]):
            if i % 2 == 0:
                traceven[even] = traces[i]
                even = even + 1
            if (i + 1) % 2 == 0:
                traceodd[odd] = traces[i]
                odd = odd + 1
        return traceodd, traceven
    else:
        print("please check data's ndim")

if __name__ == '__main__':



    trace12 = np.load('D:/sidechannel/attackmeasure/traces/m1228trace3trace4.npy')


    plt.plot(ttest(split_func(trace12)[0],split_func(trace12)[1]))

    plt.show()



import numpy as np
import estraces as traces

import filetype


from numpy import test


def is_bytes_array(traces):
    if not isinstance(traces, np.ndarray):
        raise TypeError(f'array should be a Numpy ndarray instance, not {type(traces)}.')
    if not traces.dtype == 'uint8':
        raise ValueError(f'array should be a byte array with uint8 dtype, not {traces.dtype}.')
    return True


def get_traces_num(traces):
    # with open('parallel.npy', "rb") as f:
    #     kind = filetype.guess(f)
    #     if kind is None:
    #         print('Cannot identyfy the file type')
    #         return
    # print(kind.extension)
    # print(kind.mime)
    # print('the correct extension')
    # trace_name = project.ensure_cwp_extension(filename)
    return np.shape(traces)[0]


def get_trace_sample(trace):
    return np.size(trace)


def get_traces_sample(traces):
    return np.shape(traces)[1]


def get_traces_mean(traces):
    return np.mean(traces, axis=0)


def get_traces_abs(traces):
    return np.abs(traces)


def get_traces_square(traces):
    return np.square()

def cut_func(traces,min = None,max = None):
    if traces.ndim == 1:
        new_traces  = traces[min:max]
        return new_traces
    if traces.ndim == 2:
        new_traces = np.zeros((traces.shape[0],max-min))
        for i in range(traces.shape[0]):
            new_traces[i] = traces[i][min:max]
        return new_traces
if __name__ == '__main__':
    pass


from scipy import signal
import numpy as np
import time
from functools import lru_cache
# from enum import Enum as _Enum
# 此项为优化，暂时未用到
# from functools import lru_cache as _lru_cache
'''主要用到的滤波为低通高通带通，后期可能会扩展，使用开源signal包'''
from numba import njit, jit
'生成低通滤波'
# a = signal.lfilter()

def LowPass(trace, frequency, cutoff,  axis=-1, precision='float32'):
    '''

    :param trace: initial traces
    :param frequency:
    :param cutoff: cutoff frequency
    :param axis:
    :param precision:
    :return:
    '''
    if not isinstance(trace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if not isinstance(frequency, int) and not isinstance(frequency, float):
        raise TypeError("'frequency' should be an of int or float type.")
    if frequency <= 0:
        raise ValueError("'frequency' should be positive.")
    b, a = signal.butter(3, 2 * cutoff / frequency, 'lowpass')  # 配置滤波器 8 表示滤波器的阶数
    filtedData = signal.filtfilt(b, a, trace)
    return filtedData

def HighPass(trace, frequency, cutoff,  axis=-1, precision='float32'):
    if not isinstance(trace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if not isinstance(frequency, int) and not isinstance(frequency, float):
        raise TypeError("'frequency' should be an of int or float type.")
    if frequency <= 0:
        raise ValueError("'frequency' should be positive.")
    # for value in cutoff:
    #     if not isinstance(value, int) and not isinstance(value, float):
    #         raise TypeError("'cutoff' should be a value or a collection of values of int or float type.")
    #     if value <= 0:
    #         raise ValueError("'cutoff' value(s) should be positive.")
    #     if len(cutoff) != 1:
    #         raise ValueError("cutoff should have 1 value.")
    #     if cutoff >= frequency / 2:
    #         raise ValueError("'cutoff' should be lower than frequency/2.")

    b, a = signal.butter(3, 2 * cutoff / frequency, 'highpass')  # 配置滤波器 8 表示滤波器的阶数
    filtedData = signal.filtfilt(b, a, trace)
    return filtedData


def BandPass(trace, frequency, cutoffmin, cutoffmax, axis=-1, precision='float32'):
    if not isinstance(trace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if not isinstance(frequency, int) and not isinstance(frequency, float):
        raise TypeError("'frequency' should be an of int or float type.")
    if frequency <= 0:
        raise ValueError("'frequency' should be positive.")
    # for value:
    #     if not isinstance(value, int) and not isinstance(value, float):
    #         raise TypeError("'cutoff' should be a value or a collection of values of int or float type.")
    #     if value <= 0:
    #         raise ValueError("'cutoff' value(s) should be positive.")
    #     if len(cutoffmin) != 1:
    #         raise ValueError("cutoff should have 1 value.")
    # for value in cutoffmax:
    #     if not isinstance(value, int) and not isinstance(value, float):
    #         raise TypeError("'cutoff' should be a value or a collection of values of int or float type.")
    #     if value <= 0:
    #         raise ValueError("'cutoff' value(s) should be positive.")
    #     if len(cutoffmax) != 1:
    #         raise ValueError("cutoff should have 1 value.")

    b, a = signal.butter(3, [2 * cutoffmin / frequency, 2*cutoffmax/frequency],'bandpass')
    filtedData = signal.filtfilt(b, a, trace)
    return trace
# @jit
def fast_lowpass(traces,a):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'trace' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0],traces.shape[1]), dtype=float)
    # point_list = []
    for t in range(traces.shape[0]):
        for i in range(0,traces.shape[1]):
            if i == 0:
                new_traces[t][0] = traces[t][0]
            else:
                new_traces[t][i] = (a*traces[t][i-1]+traces[t][i])/(1+a)
    return new_traces
if __name__ == '__main__':
    trace = np.load('G:/py+idea/python/sidechannel/pretrace/mtraces/mevenodd.npy')
    s_time = time.time()

    print(fast_lowpass(trace, 1))
    e_time = time.time()
    print(e_time - s_time)












from scipy import signal
import numpy as np
from tqdm import tqdm, trange
import numba as nb

np.seterr(divide='ignore', invalid='ignore')


@nb.jit
def to_one2(traces):
    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0], traces.shape[1]), dtype=float)
    a = np.mean(traces, axis=0)
    for t in range(traces.shape[0]):

        # s = np.std(traces[t])
        new_traces[t] = (traces[t] - a)
    return new_traces

# @cuda.jit
def LowPass(trace, frequency, cutoff, axis=-1, precision='float32'):
    if not isinstance(trace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if not isinstance(frequency, int) and not isinstance(frequency, float):
        raise TypeError("'frequency' should be an of int or float type.")
    if frequency <= 0:
        raise ValueError("'frequency' should be positive.")
    b, a = signal.butter(3, 2 * cutoff / frequency, 'lowpass')  # 配置滤波器 8 表示滤波器的阶数
    filtedData = signal.filtfilt(b, a, trace)
    return filtedData



@nb.jit
def moving_resamples(traces, k):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0], traces.shape[1] // k), dtype=float)
    if k >= traces.shape[1]:
        raise ValueError('window is too bigooo!')
    n = traces.shape[1]
    for t in range(traces.shape[0]):
        list = []
        # //地板除
        for i in range(n // k):
            sum = 0
            for j in range(i * k, i * k + k):
                sum = traces[t][j] + sum
            a = sum / k
            list.append(a)
        new_traces[t] = np.array(list)
    return new_traces



if __name__ == '__main__':

    # 文件路径
    url_trace = r"H:/zhoujiayun3/"
    # 原始文件名
    trace_name = ""
    # 要保存的文件名
    pre_trace_name = ""
    # 板子采样率
    sample_rate = 12e6
    # 低通滤波值
    filter_value = 2e6
    # 文件块数
    num_file_blocks = 100


    for j in trange(num_file_blocks):
        arr = np.load(url_trace+trace_name+r"{0}.npy".format(j))
        tracetoone = moving_resamples(arr, 10)
        tracepass1 = to_one2(tracetoone)
        # 此为低通滤波
        tracepass = LowPass(tracepass1, sample_rate, filter_value)
        np.save(url_trace+pre_trace_name+r'{0}.npy'.format(j), tracepass)



import numpy as np
from new_math import mean_func, cov_func
from scipy.stats import multivariate_normal

from sidechannelattack.basetrace.data_resolve import HW


# model_data应该为对应trace的中间值，故事先用中间值函数计算出来
def find_pois(model_data, model_traces, attack_data, attack_traces, num_pois, poi_space):
    """
    :param model_data: 用来建模的曲线集对应的中间值
    :param model_traces: 用来建模的曲线集
    :param attack_data: 模板攻击用到的曲线对应的 中间值
    :param attack_traces: 模板攻击用到的曲线
    :param num_pois: 设定好一条曲线有几个泄漏点
    :param poi_space: 泄漏点周边范围内设置为0
    :return: 攻击出来的值
    """
    temp_trace_hw = [[] for _ in range(9)]
    temp_means = np.zeros((9, model_traces.shape[1]))
    temp_sum_diff = np.zeros(model_traces.shape[1])
    poi_lists = []
    # Fill them up
    for i in range(model_traces.shape[0]):
        HW_number = int(model_data[i])
        temp_trace_hw[HW_number].append(model_traces[i])
    # Switch to numpy arrays
    temp_trace_hw = [np.array(temp_trace_hw[HW_number]) for HW_number in range(9)]
    for i in range(9):
        temp_means[i] = mean_func(temp_trace_hw[i])
        for j in range(i):
            temp_sum_diff += np.abs(temp_means[i] - temp_means[j])
    # 5: Find POIs
    for i in range(num_pois):
        # Find the max
        next_poi = temp_sum_diff.argmax()
        poi_lists.append(next_poi)
        # Make sure we don't pick a nearby value
        poi_min = max(0, next_poi - poi_space)
        poi_max = min(next_poi + poi_space, len(temp_sum_diff))
        for j in range(poi_min, poi_max):
            temp_sum_diff[j] = 0
    mean_matrix = np.zeros((9, num_pois))
    cov_matrix = np.zeros((9, num_pois, num_pois))
    for HW_number in range(9):
        for i in range(num_pois):
            # Fill in mean
            mean_matrix[HW_number][i] = temp_means[HW_number][poi_lists[i]]
            for j in range(num_pois):
                x = temp_trace_hw[HW_number][:, poi_lists[i]]
                y = temp_trace_hw[HW_number][:, poi_lists[j]]
                cov_matrix[HW_number, i, j] = cov_func(x, y)
    P_k = np.zeros(256)
    for j in range(attack_traces.shape[0]):
        a = [attack_traces[j][poi_lists[i]] for i in range(len(poi_lists))]
        # Test each key
        for k in range(256):
            # 此处中间值为data
            HW_number = HW[attack_data[j]]
            rv = multivariate_normal(mean_matrix[HW_number], cov_matrix[HW_number])
            p_kj = rv.logpdf(a)
            P_k[k] += p_kj
    guess = P_k.argsort()[-1]
    print(hex(guess))


if __name__ == '__main__':
    data1 = HW(np.load(r"G:/py+idea/python/sidechannel/attackmeasure/traces/SendData.npy"))

    # print(HW(data1))
    trace = np.load(r"G:/py+idea/python/sidechannel/pretrace/mtraces/m0x01evenodd.npy")
    print(find_pois(data1, trace, 5, 4))



# import numpy as np
#
#
# def pca(traces, k):
#     n_samples, n_features = traces.shape
#     print(n_samples)
#     mean = np.array([np.mean(traces[:, i]) for i in range(n_features)])
#     norm_traces = traces - mean
#     scatter_matrix = np.dot(np.transpose(norm_traces), norm_traces)
#     eig_val, eig_vec = np.linalg.eig(scatter_matrix)
#     eig_pairs = [(np.abs(eig_val[i]), eig_vec[:, i]) for i in range(n_features)]
#     eig_pairs.sort(reverse=True)
#     feature = np.array([ele[1] for ele in eig_pairs[:k]])
#     data = np.dot(norm_traces, np.transpose(feature))
#     return data
import time

from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
import numpy as np


def pr_co_an(traces, k: int):
    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if k <= 0:
        raise ValueError("k can't <0")
    if k >= traces.shape[1]:
        raise ValueError("k is toooo big!!!")
    pca = PCA(n_components=k)
    pca.fit(traces)
    traces_new = pca.transform(traces)

    return np.array(traces_new)


if __name__ == '__main__':
    n = np.load('G:/py+idea/python/sidechannel/pretrace/aes_traces.npy')

    plt.plot(n.T)
    plt.show()
    print(pr_co_an(n, 1))
    print(n.shape)



import numpy as np
import time

# np.nanmax:排除nan值求最大
np.seterr(divide='ignore', invalid='ignore')
from numba import njit, jit


# @jit
def correlation_func(data, traces):
    """

    :param data: 用于参与计算的中间值
    :param traces: 用于参与计算的曲线集
    :return: data和traces每一列的相关性
    """
    if data.ndim == 1 and traces.ndim == 2:
        old_cov = np.zeros((traces.shape[0], traces.shape[1]))
        new_cov = np.zeros((traces.shape[0], traces.shape[1]))
        old_mean_data = np.zeros(len(data))
        new_mean_data = np.zeros(len(data))
        old_mean_traces = np.zeros((traces.shape[0], traces.shape[1]))
        new_mean_traces = np.zeros((traces.shape[0], traces.shape[1]))
        old_var_data = np.zeros(len(data))
        new_var_data = np.zeros(len(data))
        old_var_traces = np.zeros((traces.shape[0], traces.shape[1]))
        new_var_traces = np.zeros((traces.shape[0], traces.shape[1]))
        for i in range(traces.shape[0] - 1):
            old_cov[0] = np.zeros(traces.shape[1])
            old_mean_data[0] = data[0]
            old_var_data[0] = 0
            old_var_traces[0] = np.zeros(traces.shape[1])
            old_mean_traces[0] = traces[0]
            new_mean_data[i + 1] = old_mean_data[i] + (data[i] - old_mean_data[i]) / (i + 1)
            new_mean_traces[i + 1] = old_mean_traces[i] + (traces[i] - old_mean_traces[i]) / (i + 1)
            old_mean_data[i + 1] = new_mean_data[i + 1]
            old_mean_traces[i + 1] = new_mean_traces[i + 1]
            new_var_data[i + 1] = old_var_data[i] + (
                    (data[i + 1] - old_mean_data[i]) * (data[i + 1] - new_mean_data[i + 1]) - old_var_data[i]) / (
                                          i + 1)
            new_var_traces[i + 1] = old_var_traces[i] + (
                    (traces[i + 1] - old_mean_traces[i]) * (traces[i + 1] - new_mean_traces[i + 1]) -
                    old_var_traces[i]) / (i + 1)
            new_cov[i + 1] = (old_cov[i] * i + (data[i + 1] - old_mean_data[i]) * (
                    traces[i + 1] - new_mean_traces[i + 1])) / (i + 1)

            old_cov[i + 1] = new_cov[i + 1]
            old_var_traces[i + 1] = new_var_traces[i + 1]
            old_var_data[i + 1] = new_var_data[i + 1]
        return np.absolute(new_cov[i + 1] / np.sqrt(new_var_traces[i + 1] * new_var_data[i + 1]))


# def correction(self, data, traces):
#     if traces.shape[0] != data.shape[0]:
#         raise ValueError("they should have same imediate value.")
#     if not isinstance(traces, np.ndarray):
#         raise TypeError("'data' should be a numpy ndarray.")
#     if not isinstance(data, np.ndarray):
#         raise TypeError("'data' should be a numpy ndarray.")
#     d_sum = 0.0
#     d2_sum = 0.0
#     t_sum = np.zeros(traces.shape[1])
#     t2_sum = np.zeros(traces.shape[1])
#     dt_sum = np.zeros(traces.shape[1])
#     tempt_sum = np.zeros(traces.shape[1])
#     tempdt_sum = np.zeros(traces.shape[1])
#     tempt2_sum = np.zeros(traces.shape[1])
#     correct_array = []
#     D = len(data)
#     n = 0
#
#     for i in range(traces.shape[0]):
#         n += 1
#
#         t_sum = tempt_sum * (n - 1) + traces[i, :]
#         tempt_sum = t_sum/n
#         t2_sum = tempt2_sum * (n-1)+(traces ** 2)[i, :]
#         tempt2_sum = t2_sum/n
#         dt_sum = tempdt_sum * (n-1)+(data[i] * traces[i])
#         tempdt_sum = dt_sum/n
#     # temp1 = 0
#     temp2 = 0
#     temp3 = 0
#     d_sum = np.sum(data)
#     d2_sum = np.sum(data**2)
#     temp1 = np.sqrt(D * d2_sum - d_sum ** 2)
#     temp2 = np.sqrt(D * t2_sum - t_sum ** 2)
#     temp3 = D * dt_sum - d_sum * t_sum
#     # temp1 = np.sqrt(var_func(data))
#     # temp2 = np.sqrt(var_func(traces))
#     # temp1 = np.sqrt(np.var(data))
#     # temp2 = np.sqrt(np.var(traces,axis=0))
#     correct_array = temp3/(temp1*temp2)
#     # print(temp1*temp2)
#     return correct_array
def one_key_guess(data, traces):
    one_key_guess_matrix = []
    for i in range(data.shape[1]):
        n = correlation_func(data[:, i], traces)
        one_key_guess_matrix.append(n)
    m = np.max(one_key_guess_matrix)
    t = np.where(one_key_guess_matrix == m)
    return t[0]


# np.argmax   np.where  np.max
if __name__ == '__main__':
    from numpy import loadtxt
    import math
    import matplotlib.pyplot as plt

    data = np.load(r"G:/py+idea/python/sidechannel/attackmeasure/traces/SendData.npy")
    trace = np.load(r"G:/py+idea/python/sidechannel/pretrace/mtraces/m0x01evenodd.npy")
    # print(data,trace)
    s_time = time.time()
    print(correlation_func(data, trace))

    e_time = time.time()
    print(e_time - s_time)
    plt.plot(correlation_func(data, trace))
    plt.show()



# 基本示例
def online_mean(old_mean, new_x, N):
    new_mean = old_mean + (new_x - old_mean) / (N + 1)
    return new_mean
def welford(old_var, old_mean, new_x, new_mean):
    new_var = old_var + ((new_x - old_mean) * (new_x - new_mean) - old_var) / (N + 1)
    return new_var
# def tt_mean(traces):
#     # 都是按列的
#     t_sum = np.zeros(traces.shape[1])
#     t_num = traces.shape[0]
#     aver_mean = np.zeros(traces.shape[1])
#     n = 0
#     for t in range(t_num):
#         n = n+1
#         t_sum = aver_mean*(n-1)+traces[t,:]
#         aver_mean = t_sum/n
#     return aver_mean
# def tt_var(traces):
#     return tt_mean(traces**2)-(tt_mean(traces))**2
# def calc_corr(a, b):
#     a_avg = sum(a) / len(a)
#     b_avg = sum(b) / len(b)
#     # 计算分子，协方差————按照协方差公式，本来要除以n的，由于在相关系数中上下同时约去了n，于是可以不除以n
#     cov_ab = sum([(x - a_avg) * (y - b_avg) for x, y in zip(a, b)])
#     # 计算分母，方差乘积————方差本来也要除以n，在相关系数中上下同时约去了n，于是可以不除以n
#     sq = math.sqrt(sum([(x - a_avg) ** 2 for x in a]) * sum([(x - b_avg) ** 2 for x in b]))
#     corr_factor = cov_ab / sq
#     return corr_factor
# def cor(data,trace1):
#     list = []
#     for t in range(trace1.shape[1]):
#         a = calc_corr(data,trace1[:,t])
#         list.append(a)
#     return np.array(list)
import numpy as np
from numpy import loadtxt
# 读入曲线

# 求矩阵均值按列
def mean_func(traces):
    # n = 0
    if traces.ndim == 1:
        old_mean = np.zeros(len(traces))
        new_mean = np.zeros(len(traces))
        for i in range(len(traces)-1):
            old_mean[0] = traces[0]
            new_mean[i + 1] = old_mean[i] + (traces[i] - old_mean[i]) / (i + 1)
            old_mean[i + 1] = new_mean[i + 1]
        return new_mean[i + 1]
    if traces.ndim ==2:
        old_mean = np.zeros((traces.shape[0],traces.shape[1]))
        new_mean = np.zeros((traces.shape[0], traces.shape[1]))
        for i in range(traces.shape[0]-1):
            # n = n+1
            old_mean[0] = traces[0]
            new_mean[i+1] = old_mean[i] + (traces[i] - old_mean[i]) / (i + 1)
            old_mean[i+1] = new_mean[i+1]
        return new_mean[i+1]

# 求矩阵方差，按行
def var_func(traces):
    if traces.ndim == 1:
        old_var = np.zeros(len(traces))
        new_var = np.zeros(len(traces))
        old_mean = np.zeros(len(traces))
        new_mean = np.zeros(len(traces))
        for i in range(len(traces) - 1):
            # 初始化
            old_var[0] = 0
            old_mean[0] = traces[0]

            # 需要的均值
            new_mean[i + 1] = old_mean[i] + (traces[i] - old_mean[i]) / (i + 1)

            # 均值的更新
            old_mean[i + 1] = new_mean[i + 1]


            new_var[i + 1] = old_var[i] + ((traces[i + 1] - old_mean[i]) * (traces[i + 1] - new_mean[i + 1]) - old_var[i]) / (i + 1)
            old_var[i + 1] = new_var[i + 1]
        return new_var[i + 1]
    if traces.ndim == 2:
        old_var = np.zeros((traces.shape[0], traces.shape[1]))
        new_var = np.zeros((traces.shape[0], traces.shape[1]))
        old_mean = np.zeros((traces.shape[0], traces.shape[1]))
        new_mean = np.zeros((traces.shape[0], traces.shape[1]))
        for i in range(traces.shape[0]-1):
            old_var[0] = np.zeros(traces.shape[1])
            old_mean[0] = traces[0]
            new_mean[i + 1] = old_mean[i] + (traces[i] - old_mean[i]) / (i + 1)
            old_mean[i + 1] = new_mean[i + 1]
            new_var[i+1] = old_var[i] + ((traces[i+1] - old_mean[i]) * (traces[i+1] - new_mean[i+1]) - old_var[i]) / (i + 1)
            old_var[i+1] = new_var[i+1]
        return new_var[i+1]


# 求一条data和trace每列的协方差
def cov_func(data,traces):
    if data.ndim == 1:
        old_cov = np.zeros((traces.shape[0], traces.shape[1]))
        new_cov = np.zeros((traces.shape[0], traces.shape[1]))
        old_mean_data = np.zeros(len(data))
        new_mean_data = np.zeros(len(data))
        old_mean_traces = np.zeros((traces.shape[0], traces.shape[1]))
        new_mean_traces = np.zeros((traces.shape[0], traces.shape[1]))

        for i in range(traces.shape[0]-1):
            # 初始化
            old_cov[0] = np.zeros(traces.shape[1])
            old_mean_data[0] = data[0]
            old_mean_traces[0] = traces[0]

            # 需要的均值
            new_mean_data[i + 1] = old_mean_data[i] + (data[i] - old_mean_data[i]) / (i + 1)
            new_mean_traces[i + 1] = old_mean_traces[i] + (traces[i] - old_mean_traces[i]) / (i + 1)

            # 均值的更新
            old_mean_data[i + 1] = new_mean_data[i + 1]
            old_mean_traces[i + 1] = new_mean_traces[i + 1]
            # 注意这里的i值
            # 协方差的计算
            new_cov[i+1] = (old_cov[i]*i + (data[i+1]-old_mean_data[i])*(traces[i+1]-new_mean_traces[i+1]))/(i+1)

            # print(old_cov[i])
            # 协方差更新
            old_cov[i+1] = new_cov[i+1]
            # print(old_cov[i+1],data[i+1]-old_mean_data[i])
        return new_cov[i+1]




# 求一条data和trace每列的相关系数
def correlation_func(data,traces):
    if data.ndim == 1:
        old_cov = np.zeros((traces.shape[0], traces.shape[1]))
        new_cov = np.zeros((traces.shape[0], traces.shape[1]))
        old_mean_data = np.zeros(len(data))
        new_mean_data = np.zeros(len(data))
        old_mean_traces = np.zeros((traces.shape[0], traces.shape[1]))
        new_mean_traces = np.zeros((traces.shape[0], traces.shape[1]))
        old_var_data = np.zeros(len(data))
        new_var_data = np.zeros(len(data))
        old_var_traces = np.zeros((traces.shape[0], traces.shape[1]))
        new_var_traces = np.zeros((traces.shape[0], traces.shape[1]))
        for i in range(traces.shape[0]-1):
            old_cov[0] = np.zeros(traces.shape[1])
            old_mean_data[0] = data[0]
            old_var_data[0] = 0
            old_var_traces[0] = np.zeros(traces.shape[1])
            old_mean_traces[0] = traces[0]
            new_mean_data[i + 1] = old_mean_data[i] + (data[i] - old_mean_data[i]) / (i + 1)
            new_mean_traces[i + 1] = old_mean_traces[i] + (traces[i] - old_mean_traces[i]) / (i + 1)
            old_mean_data[i + 1] = new_mean_data[i + 1]
            old_mean_traces[i + 1] = new_mean_traces[i + 1]
            new_var_data[i + 1] = old_var_data[i] + (
                    (data[i + 1] - old_mean_data[i]) * (data[i + 1] - new_mean_data[i + 1]) - old_var_data[i]) / (i + 1)
            new_var_traces[i + 1] = old_var_traces[i] + (
                        (traces[i + 1] - old_mean_traces[i]) * (traces[i + 1] - new_mean_traces[i + 1]) - old_var_traces[i]) / (i + 1)
            new_cov[i+1] = (old_cov[i]*i + (data[i+1]-old_mean_data[i])*(traces[i+1]-new_mean_traces[i+1]))/(i+1)

            old_cov[i + 1] = new_cov[ i + 1]
            old_var_traces[i + 1] = new_var_traces[i + 1]
            old_var_data[i + 1] = new_var_data[i + 1]
        # print(new_cov[i+1],new_var_traces[i + 1],new_var_data[i + 1])
        return new_cov[i+1]/np.sqrt(new_var_traces[i + 1]*new_var_data[i + 1])
if __name__ == '__main__':
    trace = loadtxt(r"D:\allData\FixedInput.csv",delimiter=',')



import numpy as np
from new_math import var_func, mean_func
from  matplotlib import pyplot as plt
from tqdm import tqdm, trange
import numba as nb
def block_ttest(n,url_trace):
    '''

    :param n: 因数据量大  分块数
    :param Na: 序号奇数的曲线数量
    :param Nb: 序号偶数的曲线数量
    :param N: 每条曲线的采样点
    :return:奇偶曲线的ttest结果
    '''
    # N= 2
    count = 0
    arr = np.load(url_trace + "process_arrPart0.npy")
    N = arr.shape[1]
    old_var_even = np.zeros(N)
    old_mean_even = np.zeros(N)
    old_var_odd = np.zeros(N)
    old_mean_odd = np.zeros(N)
    oddcount = 0
    evencount = 0
    for j in trange(n):
        # arr = np.load(r"D:/sidechannel/attackmeasure/1228trace1trace2/fast_lowpassarrPart{0}.npy".format(j))
        # arr = np.load('D:/sidechannel/attackmeasure/traces/temp.npy')


        # !!!!!!!!!
        # arr = np.load(r"D:/1222order1EditAlphaSboxMixAlphais1trace1trace2/yiwanarrPart{0}.npy".format(j))
        arr = np.load(r"G:/3.28/process_arrPart{0}.npy".format(j))
        # arr = np.load(r'G:/code1222order1trace1trace2alpha_sboxandmixalphaareuniqueandgoodhamming/passarrPart{0}.npy'.format(j))
        for i in range(arr.shape[0]):
            if count % 2 == 0:
                new_mean = old_mean_even + (arr[i] - old_mean_even) / (evencount + 1)
                new_var = old_var_even + ((arr[i] - old_mean_even) * (arr[i] - new_mean) - old_var_even) / (evencount + 1)
                old_mean_even = new_mean
                old_var_even = new_var
                evencount += 1
                count = count + 1
            else:
                new_mean = old_mean_odd + (arr[i] - old_mean_odd) / (oddcount + 1)
                new_var = old_var_odd + ((arr[i] - old_mean_odd) * (arr[i] - new_mean) - old_var_odd) / (oddcount + 1)
                old_mean_odd = new_mean
                old_var_odd = new_var
                oddcount += 1
                count = count + 1

    temp1 = old_mean_even-old_mean_odd
    # temp1 = old_mean_odd - old_mean_even
    temp2 = (old_var_even/evencount)+(old_var_odd/oddcount)
    test_result = temp1 / np.sqrt(temp2)
    return test_result



def block_average(n,N):
    count = 0
    old_mean_even = np.zeros(N)
    for j in trange(n):
        arr = np.load(r"G:/order1EditAll1/passarrPart{0}.npy".format(j))
        for i in range(arr.shape[0]):
            new_mean = old_mean_even + (arr[i] - old_mean_even) / (count + 1)
            old_mean_even = new_mean
            count = count + 1
    return old_mean_even/2




@nb.jit
def fast_lowpass(traces,a):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'trace' should be a numpy ndarray.")

    new_traces = np.zeros((traces.shape[0],traces.shape[1]), dtype=float)
    # point_list = []
    for t in range(traces.shape[0]):
        for i in range(0,traces.shape[1]):
            if i == 0:
                new_traces[t][0] = traces[t][0]
            else:
                new_traces[t][i] = (a*traces[t][i-1]+traces[t][i])/(1+a)
    return new_traces
# @nb.jit
# def fast_lowpass(traces,a):
#     new_traces = np.zeros(len(traces), dtype=float)
#     for i in range(0, len(traces)):
#         if i == 0:
#             new_traces[0] = traces[0]
#         else:
#             new_traces[i] = (a * traces[i - 1] + traces[i]) / (1 + a)
#     return new_traces
def ttest(t1,t2):
    Na = t1.shape[0]
    Nb = t2.shape[0]
    temp1 = mean_func(t1)-mean_func(t2)
    temp2 = (var_func(t1) / Na) + (var_func(t2) / Nb)
    test_result = temp1/np.sqrt(temp2)
    return test_result
@nb.jit
def moving_resamples(traces, k):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0], traces.shape[1] // k), dtype=float)
    if k >= traces.shape[1]:
        raise ValueError('window is too bigooo!')
    n = traces.shape[1]
    for t in range(traces.shape[0]):
        list = []
        # //地板除
        for i in range(n // k):
            sum = 0
            for j in range(i * k, i * k + k):
                sum = traces[t][j] + sum
            a = sum / k
            list.append(a)
        new_traces[t] = np.array(list)
    return new_traces
@nb.jit
def to_one2(traces):
    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0],traces.shape[1]), dtype=float)
    for t in range(traces.shape[0]):
        a = np.mean(traces[t])
        # s = np.std(traces[t])
        new_traces[t] = (traces[t]-a)
    return new_traces
if __name__ == '__main__':

    #xiangliang1 = fast_lowpass(xiangliang, 50)
    plt.rcParams['figure.figsize'] = (12.0, 4.0)
    f, ax = plt.subplots(1, 1)
    #xiangliang是一条曲线，最终结果，是对所有曲线按照列做mean
    #注意block_average里的路径要和block_ttest里的路径是一样的
    # xiangliang= block_average(200, 21000)
    # arr1 = np.load(r'd:/process_arrPart0.npy')
    # arr2 = np.load(r'd:/process_arrPart1.npy')


    # np.save(r'G:/trace/process_arrPart{0}.npy'.format(j),tracepass1)
    #c = np.r_[arr1, arr2]

    #final = np.mean(c,axis=0)
    #plt.plot(final[0:-700])
    # xiangliangt = ttest(arr1,arr2)
    #ttest结果
    xiangliangt = block_ttest(100)
    # np.save(r'D:/sidechannel/result/order1EditAll1.npy', xiangliangt)
    # plt.plot(xiangliang)
    # xiangliang = np.load(r"D:/sidechannel/result/bestorder2trace12.npy")

    # plt.plot(xiangliang[0:-750]/120)
    plt.plot(xiangliangt)
    plt.grid()
    ax.axhline(y=4.5, ls='--', c='red')
    ax.axhline(y=-4.5, ls='--', c='red')
    # plt.plot(xiangliang[0:20200])
    # plt.axhline(y=4.5, ls='--', c='red', linewidth=4)
    # plt.axhline(y=-4.5, ls='--', c='red', linewidth=4)
    plt.show()


# from . import base
# from . import filter
# # from . import aligner
# from . import fft
# from . import find_peaks
# from . import moving_average
# from . import principal_component_analysis
# from . import pre_traces




import numpy as np
from tqdm import trange


def hammingWeight(n):
    """
    :type n: int
    :rtype: int
    """
    ans = 0
    if n == 0:
        return 0
    while n > 0:
        n = n & (n - 1)
        ans += 1
    return ans


sbox = (
    0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
    0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
    0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
    0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
    0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
    0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
    0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
    0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
    0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
    0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
    0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
    0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
    0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
    0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
    0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
    0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16)


def intermediate(pt, keyguess):
    return (sbox[int(pt) ^ keyguess])


def dpa_attack(cipher_data, traces):
    mean_diffs = np.zeros(256)
    key_guess = []
    for subkey in trange(cipher_data.shape[1]):
        for kguess in range(256):
            one_list = []
            zero_list = []
            for i in range(traces.shape[0]):
                if (intermediate(cipher_data[i][subkey], kguess)& 1):
                    one_list.append(traces[i])
                else:
                    zero_list.append(traces[i])
            one_avg = np.asarray(one_list).mean(axis=0)
            zero_avg = np.asarray(zero_list).mean(axis=0)
            mean_diffs[kguess] = np.max(abs(one_avg - zero_avg))
        guess = np.argsort(mean_diffs)[-1]
        print(guess)
        key_guess.append(guess)

    return key_guess


if __name__ == '__main__':
    new = np.load(r"AES_DPA_ATTACK_textin_array.npy")[0:2000]
    a = np.load(r"AES_DPA_ATTACK_trace.npy")[0:2000]
    print(dpa_attack(new, a))
    # data = np.load(r"AES_DPA_ATTACK_textin_array.npy")
    print(np.load(r"AES_DPA_ATTACK_known_keys.npy")[0])
    # print(data.shape)



import numpy as np
from numba import cuda, float32,jit,vectorize
import time


# def moving_average(self,trace,window):
#     if not isinstance(trace, np.ndarray):
#         raise TypeError(f"'data' should be a numpy ndarray, not {type(data)}.")
#     if window >= len(trace):
#         raise ValueError('window is too bigooo!')
#     n = len(trace)
#     list = []
#     for i in range(n-window):
#         sum = 0
#         for j in range(i,i+window):
#             sum = trace[j]+sum
#         a = sum/window
#         list.append(a)
#     return list
# @vectorize(["float64(float64, float64)"], target='cuda')
def moving_average(traces, window):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'data' should be a numpy ndarray.")
    new_traces = []
    for t in range(traces.shape[0]):
        list = []
        new_traces.append(list)
        for i in range(traces.shape[1] - window):
            sum = 0
            for j in range(i, i + window):
                sum = traces[t][j] + sum
            a = sum / window
            list.append(a)
    return np.array(new_traces)


# @jit
def moving_resamples(traces, k):
    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0], traces.shape[1] // k), dtype=float)
    if k >= traces.shape[1]:
        raise ValueError('window is too bigooo!')
    n = traces.shape[1]
    for t in range(traces.shape[0]):
        list = []
        # //地板除
        for i in range(n // k):
            sum = 0
            for j in range(i * k, i * k + k):
                sum = traces[t][j] + sum
            a = sum / k
            list.append(a)
        new_traces[t] = np.array(list)
    return new_traces


if __name__ == '__main__':
    start_cpu = time.time()
    trace = np.load(r"G:/pycharm/sidechannel/attackmeasure/traces/finalBeforeOnlineSboxusingPico.npy")
    trace1 = moving_average[2,2](trace, 400)
    end_cpu = time.time()
    time_cpu = (end_cpu - start_cpu)
    print("CPU process time: " + str(time_cpu))





from scipy import signal
import numpy as np
from fastdtw import fastdtw
# from base import get_trace_sample
from find_peaks import *
import scipy.stats as stats
import matplotlib.pyplot as plt
# from attackmeasure.new_math import var_func,mean_func,correlation_func

# # 在噪声加入的情况下表现不好，但可以同步曲线  目前是两个曲线要等长
# def align_traces(ref, trace):
#     N = trace.shape[1]
#     r = 1
#     aref = np.array(list(ref))
#     atrace = np.array(list(trace))
#     dist, path = fastdtw(aref, atrace, radius=r, dist=None)
#     px = [x for x, y in path]
#     py = [y for x, y in path]
#     n = [0] * N
#     s = [0.0] * N
#     for x, y in path:
#         s[x] += trace[y]
#         n[x] += 1
#
#     ret = [s[i] / n[i] for i in range(N)]
#     return ret
def correlation_func(data,traces):
    if data.ndim == 1:
        old_cov = np.zeros((traces.shape[0], traces.shape[1]))
        new_cov = np.zeros((traces.shape[0], traces.shape[1]))
        old_mean_data = np.zeros(len(data))
        new_mean_data = np.zeros(len(data))
        old_mean_traces = np.zeros((traces.shape[0], traces.shape[1]))
        new_mean_traces = np.zeros((traces.shape[0], traces.shape[1]))
        old_var_data = np.zeros(len(data))
        new_var_data = np.zeros(len(data))
        old_var_traces = np.zeros((traces.shape[0], traces.shape[1]))
        new_var_traces = np.zeros((traces.shape[0], traces.shape[1]))
        for i in range(traces.shape[0]-1):
            old_cov[0] = np.zeros(traces.shape[1])
            old_mean_data[0] = data[0]
            old_var_data[0] = 0
            old_var_traces[0] = np.zeros(traces.shape[1])
            old_mean_traces[0] = traces[0]
            new_mean_data[i + 1] = old_mean_data[i] + (data[i] - old_mean_data[i]) / (i + 1)
            new_mean_traces[i + 1] = old_mean_traces[i] + (traces[i] - old_mean_traces[i]) / (i + 1)
            old_mean_data[i + 1] = new_mean_data[i + 1]
            old_mean_traces[i + 1] = new_mean_traces[i + 1]
            new_var_data[i + 1] = old_var_data[i] + (
                    (data[i + 1] - old_mean_data[i]) * (data[i + 1] - new_mean_data[i + 1]) - old_var_data[i]) / (i + 1)
            new_var_traces[i + 1] = old_var_traces[i] + (
                        (traces[i + 1] - old_mean_traces[i]) * (traces[i + 1] - new_mean_traces[i + 1]) - old_var_traces[i]) / (i + 1)
            new_cov[i+1] = (old_cov[i]*i + (data[i+1]-old_mean_data[i])*(traces[i+1]-new_mean_traces[i+1]))/(i+1)

            old_cov[i + 1] = new_cov[ i + 1]
            old_var_traces[i + 1] = new_var_traces[i + 1]
            old_var_data[i + 1] = new_var_data[i + 1]
        # print(new_cov[i+1],new_var_traces[i + 1],new_var_data[i + 1])
        return new_cov[i+1]/np.sqrt(new_var_traces[i + 1]*new_var_data[i + 1])
def mean_func(traces):
    # n = 0
    if traces.ndim == 1:
        old_mean = np.zeros(len(traces))
        new_mean = np.zeros(len(traces))
        for i in range(len(traces)-1):
            old_mean[0] = traces[0]
            new_mean[i + 1] = old_mean[i] + (traces[i] - old_mean[i]) / (i + 1)
            old_mean[i + 1] = new_mean[i + 1]
        return new_mean[i + 1]
    if traces.ndim ==2:
        old_mean = np.zeros((traces.shape[0],traces.shape[1]))
        new_mean = np.zeros((traces.shape[0], traces.shape[1]))
        for i in range(traces.shape[0]-1):
            # n = n+1
            old_mean[0] = traces[0]
            new_mean[i+1] = old_mean[i] + (traces[i] - old_mean[i]) / (i + 1)
            old_mean[i+1] = new_mean[i+1]
        return new_mean[i+1]

def corelation_align(traces,initialtrace,reftrace,leftmin, rightmax):
    # initialtrace为参照曲线   othertraces为其他曲线
    if not isinstance(initialtrace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if not isinstance(reftrace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    # ref因该为另一曲线的一部分
    ref = reftrace[leftmin:rightmax]
    t = len(initialtrace)
    r = len(ref)
    correct_list = []
    # for i in range(0,t-r+1):
    #     tr_correct = correlation_func(traces[i:i+r],reftrace)
    for i in range(0,t-r+1):
        tr_correct = correlation_func(initialtrace[i:i+r],ref)
        correct_list.append(tr_correct)
    # 从左到右遇到的第一个极大相关性值   因为其中一条曲线为所有采样点，故np.argmax即可代表两条曲线偏差
    m = np.argmax(correct_list)
    n = m-leftmin
    # 初始化对其之后的曲线
    newref_trace = np.zeros(len(reftrace))
    # 因为用到均值填充
    mean_trace = mean_func(traces)
    if n==0:
        newref_trace = reftrace
    if n>0:
        newref_trace[0:n-1] = mean_trace[0:n-1]
        newref_trace[n:len(reftrace)] = reftrace[0:len(reftrace)-n]
    if n<0:
        newref_trace[len(reftrace)+n+1:len(reftrace)] = mean_trace[len(reftrace)+n+1:len(reftrace)]
        newref_trace[0:len(reftrace)+n] = reftrace[-n:len(reftrace)]
    return newref_trace
# def aligner(traces):/
    # new_traces =
if __name__ == '__main__':

    trace1 = np.load(r"G:/py+idea/python/sidechannel/attackmeasure/antrace.npy")
    t1 = trace1[0]
    t2 = trace1[1][250:500]
    print(corelation_align(trace1,trace1[0],trace1[1],250,500))
    plt.plot(corelation_align(trace1,trace1[0],trace1[1],250,500))
    plt.show()







from base import *
import numpy as np
# from .base import load
# 读取曲线
# 两种方式，第一种scared   利用estraces包，import estraces as traces,然后调用函数，第二种，open project

def open_file(filename):
    filename = ensure_extension(filename)
    proj = load(filename)
    return proj


# filepath = 'G:/py+idea/python/sidechannel/basetrace/exsample.csv'
# t1 = np.genfromtxt(filepath,delimiter=',',encoding='utf-8',skip_header=2)
# npt1 = t1[:,2]
# print(npt1)
# # print(t1)


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
'''
@Project ：sidechannelattack 
@File    ：plt_ttest.py
@Author  ：suyang
@Date    ：2022/12/6 20:03 
'''
from  matplotlib import pyplot as plt
def plt_ttest(data,length,height):
    plt.rcParams['figure.figsize'] = (length,height)
    f, ax = plt.subplots(1, 1)

    # xiangliangt = block_ttest(100)

    plt.plot(data)
    plt.grid()
    ax.axhline(y=4.5, ls='--', c='red')
    ax.axhline(y=-4.5, ls='--', c='red')

    plt.show()


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
'''
@Project ：sidechannelattack 
@File    ：__init__.py.py
@Author  ：suyang
@Date    ：2022/12/6 19:07 
'''

from __future__ import absolute_import
# from .sidechannelattack import *
name = "sidechannelattack "



import numpy as np
# import matplotlib as plt
import estraces as traces
import matplotlib.pyplot as plt

def find_peaks(trace,peak_width,thre):
    if not isinstance(trace, np.ndarray):
        raise TypeError(f"'data' should be a numpy ndarray, not {type(data)}.")
    length = len(trace)
    initial_trace = []
    for i in range(1,length-1):
        if trace[i-1]<trace[i] and trace[i]>trace[i+1] and trace[i]>thre:
            initial_trace.append(i)
        elif trace[i] == trace[i+1] and trace[i]>thre:
            initial_trace.append(i)
    num = len(initial_trace)
    num2 = 0
    second_trace = initial_trace.copy()
    for j in range(1,num):
        if initial_trace[j]-initial_trace[j-1]<peak_width:
            if initial_trace[j]>initial_trace[j-1]:
                second_trace[j-1] = 0
            else:
                second_trace[j] = 0
            # global num2
            num2 = num2+1
    subnum = num - num2
    second_trace = [i for i in second_trace if i>0]

    peak_index = []

    for i in range(len(second_trace)):
        if i==0:
            index_range = np.array(initial_trace)[np.array(initial_trace)<=second_trace[i]]
        else:
            index_range = np.array(initial_trace)[(np.array(initial_trace)<=second_trace[i])&(np.array(initial_trace)>second_trace[i-1])]
        peak_index.append(index_range[np.argmax(trace[index_range],axis=0)])
    return [subnum,peak_index]
if __name__ == '__main__':

    plt.rcParams['figure.figsize'] = [40,4]
    plt.figure()
    trace = np.load('G:/py+idea/python/sidechannel/pretrace/traces.npy')
    meantrace = np.mean(trace,axis=0)
    # plt.plot(trace.T)
    # print(meantrace)
    # plt.show()
    # plt.plot(meantrace.T)
    # plt.show()
    # find_peaks(meantrace,10,0.5)
    print(find_peaks(meantrace,10,0.05))


import numpy as np
from scipy.fftpack import fft
# import matplotlib.pyplot as plt
# Fs =10e3      # 采样频率
# f1 =390       # 信号频率1
# f2 = 2e3      # 信号频率2
# t=np.linspace(0,1,Fs)   # 生成 1s 的实践序列
# noise1 = np.random.random(10e3)      # 0-1 之间的随机噪声
# noise2 = np.random.normal(1,10,10e3)
# #产生的是一个10e3的高斯噪声点数组集合（均值为：1，标准差：10）
# y=2*np.sin(2*np.pi*f1*t)+5*np.sin(2*np.pi*f2*t)+noise2
#
# def FFT (Fs,data):
#     L = len (data)                        # 信号长度
#     N =np.power(2,np.ceil(np.log2(L)))    # 下一个最近二次幂
#     FFT_y1 = np.abs(fft(data,N))/L*2      # N点FFT 变化,但处于信号长度
#     Fre = np.arange(int(N/2))*Fs/N        # 频率坐标
#     FFT_y1 = FFT_y1[range(int(N/2))]      # 取一半
#     return Fre, FFT_y1
#
# Fre, FFT_y1 = FFT(Fs,y)
# plt.figure
# plt.plot(Fre,FFT_y1)
# plt.grid()
# plt.show()




def corelation_align(traces,initialtrace,reftrace,leftmin, rightmax):
    # initialtrace为参照曲线   othertraces为其他曲线
    if not isinstance(initialtrace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if not isinstance(reftrace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    # ref因该为另一曲线的一部分
    ref = reftrace[leftmin:rightmax]
    t = len(initialtrace)
    r = len(ref)
    correct_list = []
    # for i in range(0,t-r+1):
    #     tr_correct = correlation_func(traces[i:i+r],reftrace)
    for i in range(0,t-r):
        tr_correct = correlation_func(initialtrace[i:i+r-1],ref)
        correct_list.append(tr_correct)
    # 从左到右遇到的第一个极大相关性值   因为其中一条曲线为所有采样点，故np.argmax即可代表两条曲线偏差
    m = np.argmax(correct_list)
    n = m-leftmin
    # 初始化对其之后的曲线
    newref_trace = np.zeros(len(reftrace))
    # 因为用到均值填充
    mean_trace = mean_func(traces)
    if n==0:
        newref_trace = reftrace
    if n>0:
        newref_trace[0:n-1] = mean_trace[0:n-1]
        newref_trace[n:len(reftrace)] = reftrace[0:len(reftrace)-n]
    if n<0:
        newref_trace[len(reftrace)+n+1:len(reftrace)] = mean_trace[-n+1:len(reftrace)]
        newref_trace[0:len(reftrace)+n] = reftrace[-n:len(reftrace)]
    return newref_trace
if __name__ == '__main__':
    pass


import numpy as np
from scipy import signal
import numpy as np
from tqdm import tqdm, trange
from numba import cuda
import numba as nb
import time
from pretrace.filter import fast_lowpass,LowPass
from pretrace.moving_average import moving_average
from  matplotlib import pyplot as plt
from pretrace.new_math import mean_func,var_func
from attackmeasure.to_one import to_one2
np.seterr(divide='ignore',invalid='ignore')
# import random
def moving_average(traces, window):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'data' should be a numpy ndarray.")
    new_traces = []
    for t in range(traces.shape[0]):
        list = []
        new_traces.append(list)
        for i in range(traces.shape[1] - window):
            sum = 0
            for j in range(i, i + window):
                sum = traces[t][j] + sum
            a = sum / window
            list.append(a)
    return np.array(new_traces)
@nb.jit
def to_one2(traces):
    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0],traces.shape[1]), dtype=np.float32)
    a = np.mean(traces, axis=0)
    for t in range(traces.shape[0]):

        # s = np.std(traces[t])
        new_traces[t] = (traces[t]-a)
    return new_traces
@nb.jit
def fast_lowpass(traces,a):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'trace' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0],traces.shape[1]), dtype=np.float32)
    # point_list = []
    for t in range(traces.shape[0]):
        for i in range(0,traces.shape[1]):
            if i == 0:
                new_traces[t][0] = traces[t][0]
            else:
                new_traces[t][i] = (a*traces[t][i-1]+traces[t][i])/(1+a)
    return new_traces
# @cuda.jit
def LowPass(trace, frequency, cutoff,  axis=-1, precision='float32'):
    if not isinstance(trace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if not isinstance(frequency, int) and not isinstance(frequency, float):
        raise TypeError("'frequency' should be an of int or float type.")
    if frequency <= 0:
        raise ValueError("'frequency' should be positive.")
    b, a = signal.butter(3, 2 * cutoff / frequency, 'lowpass')  # 配置滤波器 8 表示滤波器的阶数
    filtedData = signal.filtfilt(b, a, trace)
    return filtedData
@cuda.jit
def fast_lowpass_wanggeyibu(traces, a, new_traces):
    t = cuda.threadIdx.x + cuda.blockIdx.x * cuda.blockDim.x
    gridStride = cuda.gridDim.x * cuda.blockDim.x
    for j in range(t, traces.shape[0], gridStride):
        for i in range(0, traces.shape[1]):
            if i == 0:
                new_traces[t][0] = traces[t][0]
            else:
                new_traces[t][i] = (a * traces[t][i - 1] + traces[t][i]) / (1 + a)
@nb.jit
def moving_resamples(traces, k):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0], traces.shape[1] // k), dtype=np.float32)
    if k >= traces.shape[1]:
        raise ValueError('window is too bigooo!')
    n = traces.shape[1]
    for t in range(traces.shape[0]):
        list = []
        # //地板除
        for i in range(n // k):
            sum = 0
            for j in range(i * k, i * k + k):
                sum = traces[t][j] + sum
            a = sum / k
            list.append(a)
        new_traces[t] = np.array(list)
    return new_traces
def moving_resamples2(traces, k):
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0], traces.shape[1] // k), dtype=np.float32)
    if k >= traces.shape[1]:
        raise ValueError('window is too bigooo!')
    n = traces.shape[1]
    for t in range(traces.shape[0]):
        list = []
        # //地板除
        for i in range(n // k):
            # sum = 0
            # for j in range(i * k, i * k + k):
            #     sum = traces[t][j] + sum
            # print(traces[t][i*k:i*(k+1)-1])
            a = np.mean(traces[t][i*k:i*(k+1)-1])
            list.append(a)
        new_traces[t] = np.array(list)
    return new_traces
def cut_func(traces,min = None,max = None):
    if traces.ndim == 1:
        new_traces  = traces[min:max]
        return new_traces
    if traces.ndim == 2:
        new_traces = np.zeros((traces.shape[0],max-min))
        for i in range(traces.shape[0]):
            new_traces[i] = traces[i][min:max]
        return new_traces
if __name__ == '__main__':
    # 对曲线进行处理  大量多组曲线
    for j in trange(300):
        # arr = np.load(r"D:/sidechannel/attackmeasure/1229afternoon/1229afternoonTrace1Trace2arrPart{0}.npy".format(j))
        # arr = np.load(r"G:/0103nightfixtrace3trace4/0103nightfixtrace3trace4arrPart{0}.npy".format(j+118))
        # arr = np.load(r"G:/code1222order1trace1trace2alpha_sboxandmixalphaareuniqueandgoodhamming/code1222order1trace1trace2alpha_sboxandmixalphaareuniqueandgoodhammingarrPart{0}.npy".format(j))
        arr = np.load(r"G:/io_redo_share5_edit_synchronized/io_redo_share5_edit_synchronizedarrPart{0}.npy".format(j))
        # # arr = np.load(r"D:/sidechannel/attackmeasure/1227traces3traces4/passarrPart{0}.npy".format(j))
        # # arr = np.load(r"D:/sidechannel/attackmeasure/1228trace1trace2/1228trace1trace2arrPart{0}.npy".format(j))
        # # arr = np.load(r"C:/1228trace1trace2/1229trace1trace2arrPart{0}.npy".format(j))
        tracetoone = moving_resamples(arr,10)
        tracepass1 = to_one2(tracetoone)
        tracepass =LowPass(tracepass1, 12e6,2e6)

        np.save(r'G:/io_redo_share5_edit_synchronized/process_arrPart{0}.npy'.format(j),tracepass1)
        # arr = np.load(r"D:/sidechannel/attackmeasure/1228trace1trace2/passarrPart{0}.npy".format(j))
        #
        # tracepass = fast_lowpass(arr,50)
        #
        # np.save('D:/1222order1EditAlphaSboxMixAlphais1trace1trace2/yiwanarrPart{0}.npy'.format(j), tracepass)
    # # 单独一组曲线
    # arr = np.load(r"G:/0103trace1trace2_5000tiao/0103trace1trace2.npy")
    # tracetoone = moving_resamples(arr, 50)
    # tracepass = to_one2(tracetoone)
    # np.save(r'G:/0103trace1trace2/passarrPart50.npy',tracepass)

    # print(cut_func(arr,0,10))

    # arr = np.load(r'G:/0103trace1trace2/passarrPart50.npy')
    # plt.plot(np.mean(arr,axis = 0))
    # plt.show()
    # arr = np.load(r"G:/trace/arrPart{0}.npy".format(j))
    #     # # arr = np.load(r"D:/sidechannel/attackmeasure/1227traces3traces4/passarrPart{0}.npy".format(j))
    #     # # arr = np.load(r"D:/sidechannel/attackmeasure/1228trace1trace2/1228trace1trace2arrPart{0}.npy".format(j))
    #     # # arr = np.load(r"C:/1228trace1trace2/1229trace1trace2arrPart{0}.npy".format(j))
    #     tracetoone = moving_resamples(arr,10)
    #     tracepass1 = to_one2(tracetoone)
    #     tracepass =LowPass(tracepass1, 120e6,24e6)
    #
    #     np.save(r'G:/trace/process_arrPart{0}.npy'.format(j),tracepass1)
    # arr = np.load(r"d:/arr.npy")
    # tracetoone = moving_resamples(arr, 10)
    # tracepass1 = to_one2(tracetoone)
    # tracepass =LowPass(tracepass1, 120e6,24e6)
    #
    # np.save(r'd:/process_arrPart0.npy',tracepass1)
    # # np.save(r'G:/trace/process_arrPart{0}.npy', tracepass1)
    # print(arr.shape)





import numpy as np
import estraces as traces
# 这块没有用，因为filetype包暂未支持npy文件
# import filetype
import logging
import os
import re
import sys
from datetime import datetime
import zipfile
# ths = scared.traces.read_ths_from_ets_file()
from numpy import test
# import pandas as pd
import estraces as traces


# so i need a tool to make others into npy

def ensure_extension(path):

    root, ext = os.path.splitext(path)
    if ext == '.npy':
        return path
    elif ext =='.ets':
        return traces.read_ths_from_ets_file(path).samples[:]
    elif ext == '.csv':
        # filepath = 'G:/py+idea/python/sidechannel/basetrace/exsample.csv'
        t1 = np.genfromtxt(filepath, delimiter=',', encoding='utf-8', skip_header=2)
        return t1[:,2]
        # print(t1)
    # if ext != '.npy':
    #     raise IOError("  not exist or is not a path")
    else:
        raise IOError("  not exist or is not a path")

def load(self, f = None):
    if f is not None:
        self.setFilename(os.path.abspath(os.path.expanduser(f)))

    if not os.path.isfile(self.filename):
        raise IOError("File " + self.filename + " does not exist or is not a file")

    self.config = ConfigObjProj(infile=self.filename, callback=self.configObjChanged)
    self._traceManager.loadProject(self.filename)
    self.dirty.setValue(False)


def is_bytes_array(traces):
    if not isinstance(traces, np.ndarray):
        raise TypeError(f'array should be a Numpy ndarray instance, not {type(traces)}.')
    if not traces.dtype == 'uint8':
        raise ValueError(f'array should be a byte array with uint8 dtype, not {traces.dtype}.')
    return True


def delete_one_trace(traces,order):
    # newtraces = []
    newtraces = np.delete(traces,order,axis=0)
    return newtraces

def delete_onemore_trace(traces,order1,order2):
    # newtraces = []
    newtraces = np.delete(traces,[order1,order2],axis=0)
    return newtraces



def get_traces_num(traces):
    # with open('parallel.npy', "rb") as f:
    #     kind = filetype.guess(f)
    #     if kind is None:
    #         print('Cannot identyfy the file type')
    #         return
    # print(kind.extension)
    # print(kind.mime)
    # print('the correct extension')
    # trace_name = project.ensure_cwp_extension(filename)
    return np.shape(traces)[0]

def get_trace_sample(trace):
    return np.size(trace)

def get_traces_sample(traces):
    return np.shape(traces)[1]


def get_traces_mean(traces):
    return np.mean(traces, axis=0)


def get_traces_abs(traces):
    return np.abs(traces)




import matplotlib.pyplot as plt
import numpy as np
import re
from scipy.stats import multivariate_normal
from tqdm import trange
import sklearn
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler


# numPOIs: 有多少个PoI
# POIspacing: 至少相隔多元
def choosePoI(correlation_file_path, correlation_file_name, numPOIs, POIspacing):
    corr_trace = abs(np.load(correlation_file_path + correlation_file_name))[correlation_file_offset:]
    PoIs = []
    # Repeat until we have enough POIs
    for i in range(numPOIs):
        # Find the biggest peak and add it to the list of POIs
        nextPOI = corr_trace.argmax()
        PoIs.append(nextPOI)

        # 把周围的点清零
        # Make sure we don't go out of bounds
        poiMin = max(0, nextPOI - POIspacing)
        poiMax = min(nextPOI + POIspacing, len(corr_trace))
        for j in range(poiMin, poiMax):
            corr_trace[j] = 0
    return PoIs


# 归一化
def to_one2(traces):
    # 判断是否是一个类型isinstance 不能用在numba里面
    # if not isinstance(traces, np.ndarray):
    #     raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0], traces.shape[1]), dtype=float)
    a = np.mean(traces, axis=0)
    for t in range(traces.shape[0]):
        # a = np.mean(traces, axis=0)
        new_traces[t] = (traces[t] - a)
    return new_traces


sbox = (
    0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
    0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
    0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
    0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
    0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
    0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
    0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
    0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
    0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
    0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
    0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
    0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
    0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
    0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
    0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
    0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16)


def hammingWeight(n):
    c = 0
    while n:
        c += 1
        n &= n - 1

    return c


'''
for alun in range(0, 1000):
    f1 = open("F:/tracexinzeng32sh8/data45{0}.txt".format(alun), "r")
    fdata = open("F:/tracexinzeng32sh8/aaadata{0}.txt".format(alun), "a")
    # fxiyj = open("aaaxiyj{0}.txt".format(alun), "a")
    f1list = f1.readlines()
    f1str = f1list[0]
    f1str = f1str.lstrip()
    data = []
    liststr = re.findall(r'\d+', f1str)
    for i in range(len(liststr)):
        if (i % 45) < 32:
            data.append(int(liststr[i]))
        if (i + 1) % 45 == 0:
            outputdata = data[0]

            fdata.write(str(outputdata))
            fdata.write('\n')
            data = []

    f1.close()

    fdata.close()

'''
'''

# 下面是测SNR


import numpy as np
from matplotlib import pyplot as plt
from tqdm import tqdm, trange

def prepare_data(url_trace, trace_name, url_data, data_name, n,loadData):
    m = {}
    v = {}
    #trace = np.load(url_trace + r"arrPart0.npy")
    trace = np.load(url_trace + r"arrPart0.npy")
    #count_temp = np.zeros(300, dtype=np.int32)

    # count_temp是统计每一个label的数量
    # 这里300是冗余了
    count_temp = np.zeros(300, dtype=np.int32)

    # 初始化m和v
    for i in loadData:
        m[i] = np.zeros(trace.shape[1])
        v[i] = np.zeros(trace.shape[1])
    # n是多少个block
    for j in trange(n):
        data = np.loadtxt(url_data + data_name + r"{0}.txt".format(j), delimiter=',', dtype="int")
        # print(data)
        trace = np.load(url_trace +trace_name+ r"{0}.npy".format(j))
        for count, label in enumerate(data):
            # 枚举，count表示第几个，从0开始，到499

            # d[label].append(trace[count])

            # old_mean是一个向量
            # m[label]是一个向量
            old_mean = m[label]

            # 遍历每一条数据，在线计算均值和方差 Welford's online algorithm
            # m是均值
            # v是方差
            m[label] = m[label] + (trace[count] - m[label]) / (count_temp[label] + 1)
            if label == 1:
                print("trace:",trace[count])
                print("mean:",m[label])
            # sigma的平方
            v[label] = v[label] + ((trace[count] - old_mean) * (trace[count] - m[label]) - v[label]) / (
                    count_temp[label] + 1)
            count_temp[label] += 1
    signal = []
    noise = []
    for i_ in loadData:
        # signal_trace 存所有label的均值曲线
        signal.append(m[i_])
        #print("m[i_]")
        #print(m[i_])
    for j_ in loadData:
        noise.append(v[j_])
        #print("v[j_]")
        #print(v[j_])
    signal_var = np.var(signal, axis=0)

    # 这里好像不太一样/
    noise_mean = np.mean(noise, axis=0)
    return signal_var / noise_mean

def snr_function(n,url_trace,trace_name,url_data,data_name):
    plt.rcParams['figure.figsize'] = (12.0, 4.0)
    f, ax = plt.subplots(1, 1)
    ax.set_title('snr_traces')
    list = (np.unique(np.loadtxt(url_data + data_name+r"0.txt", delimiter=',', dtype="int"))).tolist()
    for i in range(n - 1):
        dataaaa = np.loadtxt(url_data + data_name + r"{0}.txt".format(i + 1), delimiter=',', dtype="int")
        num = (np.unique(dataaaa)).tolist()
        list = list + num
    loadData = (np.unique(list))
    # 得到的loadData是*所有块中*去重后的输入数据
    result_data = prepare_data(url_trace, trace_name, url_data, data_name, n,loadData)
    np.save(r"F:/tracexinzeng32sh8/snr_data_test.npy",result_data)
    ax.plot(result_data)
    plt.show()

if __name__ == '__main__':

    # 确定文件路径，名字，块数    

    url_trace = r"F:/tracexinzeng32sh8/"
    url_data = r"F:/tracexinzeng32sh8/"

    data_name = "aaadata"
    trace_name = "arrPart"
    n = 2
    # 运行此函数
    snr_function(n,url_trace,trace_name,url_data,data_name)


# 确定的几个POI
# x = 3134 4140 5160 6170 7190 8190 9220 10220 11240 12270 13300 14280  15310 16330 17330 18350 19360 20430 21390 23400 25430 27450  29470


'''

print("loadtxt后用处理成汉明重量了！")
##########################
# 下面是路径参数 #
##########################

# file_path = r"F:/weixinzeng_wave_filter_trace_16bit/"
# file_path = r"F:/weixinzeng32sh8/"
# file_path = r"F:/another_CPU/5mhz_filter_8bit_new/"
# file_path = "F:/another_CPU/5mhz_filter_16bit/"
file_path = r"D:/ChipWhisperer5_52/cw/home/portable/chipwhisperer/tutorials/courses/sca101/example_aes/"

# file_path = r"D:/ChipWhisperer5_52/cw/home/portable/chipwhisperer/jupyter/courses/sca101/traces/"
# input_data_file_name = "aaadata{0}.txt"
input_data_file_name = "example_aes_inter_part{0}.txt"
# input_data_file_name = "lab3_3_zhongjianzhi_chaifen{0}.txt"
# traces_file_name = "arrPart{0}.npy"
traces_file_name = "example_aes_part{0}.npy"
# traces_file_name = "lab3_3_traces_chaifen{0}.npy"
# 注意要有相关性曲线的文件
# correlation_file_path = r"C:/Users/jfj/Desktop/PoI/another_CPU/5mhz_filter_16bit/"
correlation_file_path = r"C:/Users/jfj/Desktop/PoI/aes_test/"
correlation_file_name = "xiang_guan_xing.npy"
# correlation_file_name = "snr_data_better_var.npy"
correlation_file_offset = 5000  # 表示从第多少个点开始选PoI。因为第一个尖峰不看
correlation_file_offset = 0
num_of_each_part = 500  # 每个Part多少条曲线

numPOIs = 30  # 选择多少个PoI
POIspacing = 10  # PoI之间的间隔至少为多少

##########################
# 下面是建模参数 #
##########################
# 980
part = 150  # 选择多少个 part 进行建模
profiling_part_start_index = 0  # 建模曲线首先从哪一个part开始
num_of_class = 256  # 对于8bit HW是9类， 16bitHW是17类， 0-255有256类, 表示攻击的时候对多少类进行攻击
use_hammming = False
##########################
# 下面是PCA参数 #
##########################
use_pca = False  # 是否使用降维
pca_part = part  # 用多少 part 用来做 pca
pca_part_start_index = profiling_part_start_index  # 首先从哪一个part开始
pca_components = 3
# pca_method = ["full_traces", "mean_9", "mean_256", "mean_17"] 对应的下标
pca_method_select = 1

##########################

print("是否使用降维:", use_pca)

# 这个PoI对应的是 AES D:\ChipWhisperer5_52\cw\home\portable\chipwhisperer\jupyter\courses\sca101\traces  lab3_3_yihuohoudezhiPart0.npy 的脚本自动挑选出来的
# PoIs = [1313, 1318, 1341, 2165, 1661]

# full 是 对全体 pca_part * num_of_each_part 条曲线做PCA_fit
# mean_9 是用汉明重量，对9条平均曲线做PCA_fit
# mean_256 是对256条平均曲线做PCA_fit

PoIs = choosePoI(correlation_file_path, correlation_file_name, numPOIs, POIspacing)

print("PoIs", PoIs)
scale = StandardScaler()
pca_method = ["full_traces", "mean_9", "mean_256", "mean_17"]

# PoI的个数
num_of_PoIs = len(PoIs)

if use_pca == True:
    pca = PCA(n_components=pca_components)
    print("开始降维")
    if pca_method[pca_method_select] == "full_traces":
        # 暂时存放用来pca的所有曲线
        temp_for_pca = np.zeros((pca_part * num_of_each_part, num_of_PoIs))
        # 遍历part
        for i in trange(pca_part):
            # 把用来PCA的曲线都放到   temp_for_pca 里面
            t = np.load(file_path + traces_file_name.format(i + pca_part_start_index))
            t = scale.fit_transform(t)  # 标准化
            t = t[:, PoIs]  # 挑出PoI
            temp_for_pca[i * num_of_each_part:(i + 1) * num_of_each_part] = t
        # 全部放完了，做拟合
        pca.fit(temp_for_pca)
    if pca_method[pca_method_select] == "mean_9" or pca_method[pca_method_select] == "mean_17" or pca_method[
        pca_method_select] == "mean_256":
        if pca_method[pca_method_select] == "mean_9":
            pca_classes = 9
        elif pca_method[pca_method_select] == "mean_17":
            pca_classes = 17
        else:
            pca_classes = 256
        # 此时的temp_for_pca里面存的是均值, 相当于meanMatrix
        temp_for_pca = np.zeros((pca_classes, num_of_PoIs))
        # 记录每一个label所对应的曲线数量
        pca_count = [0] * pca_classes
        for i in trange(pca_part):
            label_temp = np.loadtxt(file_path + input_data_file_name.format(i + pca_part_start_index), delimiter=' ',
                                    dtype="int")

            # 如果选择的是mean_9或mean_17就计算hamming重量
            if pca_method[pca_method_select] == "mean_9" or pca_method[pca_method_select] == "mean_17":
                label_temp = [hammingWeight(label) for label in label_temp]

            trace_temp = np.load(file_path + traces_file_name.format(i + pca_part_start_index))
            trace_temp = scale.fit_transform(trace_temp)
            trace_temp = trace_temp[:, PoIs]  # 挑出PoI

            # 遍历一个Part中所有曲线
            for index, label in enumerate(label_temp):
                pca_count[label] += 1
                # # 算出新的均值
                # meanMatrix[label] = meanMatrix[label] + (trace - meanMatrix[label]) / (count[label])  # 算出新的均值
                temp_for_pca[label] = temp_for_pca[label] + (trace_temp[index] - temp_for_pca[label]) / (
                    pca_count[label])

        # 到这里计算出了均值，即temp_for_pca
        # print(temp_for_pca.shape)
        # plt.plot(temp_for_pca.T)
        # for i in range(temp_for_pca.shape[0]):
        #     plt.plot(temp_for_pca[i])
        plt.show()
        if temp_for_pca[0][0] == 0:
            print("均值PCA时剔除零值")
            print("temp_for_pca去掉第一行")
            temp_for_pca = temp_for_pca[1:]
        if temp_for_pca[-1][0] == 0:
            print("均值PCA时剔除零值")
            print("temp_for_pca去掉最后一行")
            temp_for_pca = temp_for_pca[:-1]

        # 拟合
        pca.fit(temp_for_pca)

if use_pca == False:
    # 初始化均值向量
    meanMatrix = [None] * num_of_class  # 对于每个类，存放每个PoI处的均值
    old_mean = [None] * num_of_class
    # 初始化协方差矩阵
    covMatrix = [None] * num_of_class
    # 初始化count ，用来记录每一个label下曲线的数目
    count = [None] * num_of_class
    for label in range(num_of_class):
        meanMatrix[label] = np.zeros(num_of_PoIs)
    for label in range(num_of_class):
        old_mean[label] = np.zeros(num_of_PoIs)
    for label in range(num_of_class):
        covMatrix[label] = np.zeros((num_of_PoIs, num_of_PoIs))
    for label in range(num_of_class):
        count[label] = 0

# 区别只有 num_of_class 变成pca_components
else:
    # 初始化均值向量
    meanMatrix = [None] * num_of_class  # 对于每个类，存放每个PoI处的均值
    old_mean = [None] * num_of_class
    # 初始化协方差矩阵
    covMatrix = [None] * num_of_class
    # 初始化count ，用来记录每一个label下曲线的数目
    count = [None] * num_of_class
    for label in range(num_of_class):
        meanMatrix[label] = np.zeros(pca_components)
    for label in range(num_of_class):
        old_mean[label] = np.zeros(pca_components)
    for label in range(num_of_class):
        covMatrix[label] = np.zeros((pca_components, pca_components))
    for label in range(num_of_class):
        count[label] = 0

print("开始建模")

for p in trange(part):
    input_data = np.loadtxt(file_path + input_data_file_name.format(p + profiling_part_start_index), delimiter=' ',
                            dtype="int")  # 读取一个part的输入数据
    if use_hammming:
        input_data = [hammingWeight(label) for label in input_data]  # 转化成汉明重量

    # 从profiling_part_start_index开始，读取这个 part 的曲线
    traces = np.load(file_path + traces_file_name.format(p + profiling_part_start_index))

    # traces = to_one2(traces)    # 标准化一下
    traces = scale.fit_transform(traces)
    traces = traces[:, PoIs]  # 只选取这些PoI。
    assert traces.shape[1] == len(PoIs)
    if use_pca == True:
        traces = pca.transform(traces)  # 使用PCA转化一下曲线,进一步减少特征
        assert traces.shape[1] == pca_components
    assert len(input_data) == traces.shape[0]

    # 遍历这个 part 中所有的曲线
    # online 计算 cov
    # index 是下标
    for index, label in enumerate(input_data):
        count[label] += 1  # label 对应的曲线数量+1
        trace = traces[index]  # 取出这条曲线
        old_mean[label] = meanMatrix[label]  # 暂存一下旧的均值

        # 在线计算均值和方差 Welford's online algorithm
        # m是均值
        meanMatrix[label] = meanMatrix[label] + (trace - meanMatrix[label]) / (count[label])  # 算出新的均值
        # 在线计算协方差
        for i in range(traces.shape[1]):  # 这里不用num_of_PoIs，因为如果PCA了，特征数量更少，不是 num_of_PoIs
            for j in range(traces.shape[1]):
                x = trace[i]
                y = trace[j]

                dx = x - old_mean[label][i]

                # new meanx
                # meanx += dx / count[label]

                # new meany
                # meany += (y - meany) / count[label]
                # C =( C * (n-1) +  dx * (y - meany) ) / n
                covMatrix[label][i][j] += dx * (y - meanMatrix[label][j])
# 所有 part 遍历完成之后，最终求出协方差
for label in range(num_of_class):
    for i in range(traces.shape[1]):
        for j in range(traces.shape[1]):
            covMatrix[label][i][j] = covMatrix[label][i][j] / (count[label] - 1)

# 求均值矩阵，协方差矩阵这部分验证了，没有问题

# 下面是对多条曲线进行攻击,但是每个label只使用一条曲线进行攻击，计算所有攻击的成功率
P_k = np.zeros(num_of_class)  # num_of_class 种猜测的分数
bayes_score = np.zeros(num_of_class)
##########################
# 下面是进行攻击部分的参数 #
##########################
use_bayes = True
attack_part = 3  # 攻击曲线使用多少个part

##########################
attack_part_start_index = part  # 从建模曲线的下一条曲线开始攻击

success_num = 0  # 记录有多少条攻击成功了
count_of_used_traces = 0  # 记录有多少条label为target_label的曲线被使用了
success_rate_list = []  # 记录每一次攻击的成功率
print("开始攻击，对多条曲线, 求成功率")
print("是否使用PCA", use_pca)
count_of_each_label_when_attacking = [0] * num_of_class
probability_of_each_label_when_attacking = [0] * num_of_class

for i in trange(attack_part):  # 先统计一下每个label的数目
    input_data = np.loadtxt(file_path + input_data_file_name.format(i + attack_part_start_index), delimiter=',',
                            dtype="int")
    if use_hammming:
        input_data = [hammingWeight(label) for label in input_data]

    for label in input_data:  # 遍历一个Part 中的所有输入
        count_of_each_label_when_attacking[label] += 1
test_total = 0
for label in range(num_of_class):  # 计算每一个label的概率
    probability_of_each_label_when_attacking[label] = (
                count_of_each_label_when_attacking[label] / (attack_part * num_of_each_part))
    test_total += probability_of_each_label_when_attacking[label]
print("test_total", test_total)
print("probability_of_each_label_when_attacking", probability_of_each_label_when_attacking)
for i in trange(attack_part):
    # 从 attack_part_start_index 开始，依次读取一个part的曲线
    traces = np.load(file_path + traces_file_name.format(i + attack_part_start_index))
    # traces = to_one2(traces)    # 标准化一下
    traces = scale.fit_transform(traces)
    traces = traces[:, PoIs]  # 只选取这些PoI。
    assert traces.shape[1] == len(PoIs)
    # 降维
    if use_pca:
        traces = pca.transform(traces)
        assert traces.shape[1] == pca_components
    input_data = np.loadtxt(file_path + input_data_file_name.format(i + attack_part_start_index), delimiter=',',
                            dtype="int")
    if use_hammming:
        input_data = [hammingWeight(label) for label in input_data]
    assert traces.shape[0] == len(input_data)
    for j in range(traces.shape[0]):  # 遍历这个part的每一条曲线
        count_of_used_traces += 1  # 使用的曲线数目 +1
        # a = [traces[j][PoIs[i]] for i in range(len(PoIs))]  # 取出PoI所对应的点，重组成a
        for label in range(num_of_class):  # 测试每一个模板 ，本来这里写的是 for label in input_data_unique
            # 有边界情况：协方差矩阵为nan，说明没有这个模板, 说明输入数据没有这个
            # 如果为0，说明输入数据只有1条曲线有这个，没法算出方差，方差为0
            if np.isnan(covMatrix[label][0][0]) or np.isnan(meanMatrix[label][0]) or covMatrix[label][0][0] == 0 or \
                    meanMatrix[label][0] == 0:
                print("label为" + str(label) + "的模板不存在,建模曲线数量不够")
                P_k[label] = float("-inf")  # 不存在的话，这个模板对应的分数就是负无穷
                continue
            rv = multivariate_normal(meanMatrix[label], covMatrix[label], allow_singular=True)  # 选择 label 所对应的模型
            p_kj = rv.pdf(traces[j])  # 将曲线带入公式
            # 注意 这里不是+=了,不是求累计的分数
            if use_bayes == False:
                P_k[label] = np.log(p_kj)  # 每一个模板的分数放进P_k
            else:
                P_k[label] = p_kj  # 取对数后会出现负无穷
        if use_bayes == False:  # 不使用Bayes
            # 如果最右边的数（猜测的密钥中分数最高的 所对应的下标） == label
            if P_k.argsort()[-1] == input_data[j]:  # input_data[j] 是这条曲线对应的 label，已经经过汉明重量的处理了
                success_num += 1
        else:  # 使用bayes
            denominator = 0
            for label in range(num_of_class):
                # P_k[label] 是 P(x|k)
                # probability_of_each_label_when_attacking[label]是 P(k)
                denominator += P_k[label] * probability_of_each_label_when_attacking[label]
            for label in range(num_of_class):  # 开始猜测
                bayes_score[label] = (probability_of_each_label_when_attacking[label] * P_k[label]) / denominator
            if bayes_score.argsort()[-1] == input_data[j]:  # input_data[j] 是这条曲线对应的 label，已经经过汉明重量的处理了
                success_num += 1
        print("成功率", success_num / count_of_used_traces)
        success_rate_list.append(success_num / count_of_used_traces)

np.save("success_rate.npy", success_rate_list)
plt.plot(success_rate_list)
plt.show()

'''
# 下面这个验证是正确的！！！！
# 下面是对一条曲线进行攻击
# 累加256种密钥猜测的分数
# 适用于chipwhisperer_lab3_3官方数据集用来验证用
# 原本是数据集是2500条，被我拆分成了500条一个part
attack_part = 1  # 攻击曲线使用多少个part
attack_part_start_index = 4  # 从第几个part开始

true_key  = 0x2b
P_k = np.zeros(256)

for i in trange(attack_part):
    atkTraces = np.load(file_path + traces_file_name.format(i + attack_part_start_index))
    # 取最后500个
    atkPText = np.load(r"D:/ChipWhisperer5_52/cw/home/portable/chipwhisperer/jupyter/courses/sca101/traces/lab3_3_textin.npy")[-500:]
    for j in range(len(atkTraces)):
        # Grab key points and put them in a small matrix
        a = [atkTraces[j][PoIs[i]] for i in range(len(PoIs))]

        # Test each key
        for k in range(256):
            # Find HW coming out of sbox

            HW = hammingWeight(sbox[atkPText[j][0] ^ k])
            input = np.loadtxt(file_path + input_data_file_name.format(i + attack_part_start_index), delimiter=',',
                                    dtype="int").tolist()

            if k == true_key:
                assert sbox[atkPText[j][0] ^ k] == input[j]
            # Find p_{k,j}
            rv = multivariate_normal(meanMatrix[HW], covMatrix[HW])
            p_kj = rv.pdf(a)

            # Add it to running total
            P_k[k] += np.log(p_kj)

        # Print our top 5 results so far
        # Best match on the right
        print(P_k.argsort()[-5:])

'''


import numpy as np


def to_one(traces):
    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0],traces.shape[1]), dtype=float)
    for t in range(traces.shape[0]):
        a = np.mean(traces[t])
        s = np.std(traces[t])
        new_traces[t] = (traces[t]-a)/s
    return new_traces
def to_one2(traces):
    if not isinstance(traces, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    new_traces = np.zeros((traces.shape[0],traces.shape[1]), dtype=float)
    for t in range(traces.shape[0]):
        a = np.mean(traces[t])
        # s = np.std(traces[t])
        new_traces[t] = (traces[t]-a)
    return new_traces
if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from pretrace import filter
    import random
    from numpy import loadtxt
    plt.subplot(211)
    plt.plot(to_one(loadTrace).T)
    plt.subplot(212)
    plt.plot(loadTrace.T)

    plt.show()
    print(loadTrace[0,0:10])
    print(to_one(loadTrace)[0,0:10])


import numpy as np
from tqdm import trange
import matplotlib.pyplot as plt


def t_test(n, url_trace):
    '''
    :instrctions:even  and  odd 间或采集
    :param n: blocks of traces
    :param N: sample numbers
    :return: result
    '''
    arr = np.load(url_trace + r"arrPart0.npy")
    count = 0
    N = arr.shape[1]
    old_var_even = np.zeros(N)
    old_mean_even = np.zeros(N)
    old_var_odd = np.zeros(N)
    old_mean_odd = np.zeros(N)
    oddcount = 0
    evencount = 0
    for j in trange(n):
        arr = np.load(url_trace + r"arrPart{0}.npy".format(j))
        for i in range(arr.shape[0]):
            if count % 2 == 0:
                new_mean = old_mean_even + (arr[i] - old_mean_even) / (evencount + 1)
                new_var = old_var_even + ((arr[i] - old_mean_even) * (arr[i] - new_mean) - old_var_even) / (
                        evencount + 1)
                old_mean_even = new_mean
                old_var_even = new_var
                evencount += 1
                # print(evencount)
                count = count + 1
            else:
                new_mean = old_mean_odd + (arr[i] - old_mean_odd) / (oddcount + 1)
                new_var = old_var_odd + ((arr[i] - old_mean_odd) * (arr[i] - new_mean) - old_var_odd) / (oddcount + 1)
                old_mean_odd = new_mean
                old_var_odd = new_var
                oddcount += 1
                count = count + 1
    temp1 = old_mean_even - old_mean_odd
    temp2 = (old_var_even / evencount) + (old_var_odd / oddcount)
    test_result = temp1 / np.sqrt(temp2)
    return test_result

#
# def plt_t_test(all_kinds_of_results):
#     plt.rcParams['figure.figsize'] = (12.0, 7.0)
#     f, ax = plt.subplots(1, 1)
#     plt.plot(t_test(200, url_trace))
#     ax.axhline(y=4.5, ls='--', c='red', linewidth=2)
#     ax.axhline(y=-4.5, ls='--', c='red', linewidth=2)
#     plt.show()


import numba as nb
import time


# @nb.jit
# def initial_tranform():
#     new_arr = np.zeros((45000, 7500), dtype=np.float32)
#     # for i in trange(int(1800000 / 45000)):
#     # 一共有多少块
#     m = 8
#     for i in trange(int(18000000 / 45000)):
#         for j in trange(300):
#             arr1 = np.load(r"G:/io_redo_share5_edit_synchronized/process_arrPart{0}.npy".format(j)).T
#             new_arr[:, 25 * j:25 * j + 25] = arr1[0:45000]
#         np.save(r'E:/D/new_arr{0}.npy'.format(i), new_arr)
# def transform():
#     new_arr = np.zeros((1800000, 50), dtype=np.float32)
#     # for i in trange(int(1800000 / 45000)):
#     # 一共有多少块
#     # m = 8
#     for i in trange(int(300/2)):
#         # 第一轮几块拼接
#         for j in range(2):
#             # start = time.time()
#             arr1 = np.load(r"G:/io_redo_share5_edit_synchronized/process_arrPart{0}.npy".format(j+i*2)).T
#             # arr = np.concatenate((arr1, arr2), axis=1)
#             new_arr[:, 25 * j:25 * j + 25] = arr1
#         # print(i)
#         for k in range(int(1800000 / 90000)):
#             np.save(r'D:/1/transform{0}_pre_arrPart{1}.npy'.format(i, k), new_arr[90000*k:90000*k+90000])
# def pinjie_function():
#     new_arr = np.zeros((90000, 7500), dtype=np.float32)
#     for i in trange(int(1800000/90000)):
#         for j in range(int(300/2)):
#             arr1 = np.load(r"D:/1/transform{0}_pre_arrPart{1}.npy".format(j, i))
#             # arr1 = np.load(r"E:/D/transform{0}_arrPart{1}.npy".format(j,i))
#             # arr = np.concatenate((arr1, arr2), axis=1)
#             new_arr[:, 50 * j:50 * j + 50] = arr1
#         np.save(r'D:/1/new_arr{0}.npy'.format(i), new_arr)

if __name__ == '__main__':
    # url_trace = r"H:/traceaes2/"
    # plt.rcParams['figure.figsize'] = (12.0, 7.0)
    # f, ax = plt.subplots(1, 1)
    # ax.axhline(y=4.5, ls='--', c='red', linewidth=2)
    # ax.axhline(y=-4.5, ls='--', c='red', linewidth=2)
    # result = t_test(100, url_trace)
    # # np.save(r'G:/side_channel_attack/side_channel_attack/pre_result3.npy', result )
    # # result = np.load(r'G:/side_channel_attack/result_io_redo_share5_edit_synchronized.npy')
    # # print(t_test(1,url_trace))
    # # print(np.argmax(result))
    # plt.plot(result)
    # # arr = np.load(r"H:/traceaes2/arrPart0.npy")
    # # print(arr.shape)
    # plt.show()
    # import time
    # initial_tranform()
    # transform()
    # pinjie_function()
    import gc
    import time
    start= time.time()
    arr = np.load(r"G:/io_redo_share5_edit_synchronized/process_arrPart299.npy")
    arr2 = np.load(r'D:/1/new_arr0.npy')
    # arr1 = np.load(r"G:/io_redo_share5_edit_synchronized/process_arrPart299.npy")
    print(time.time()-start)
    # print(arr[0])
    print(arr)
    print(arr.shape)
    print(arr2.shape)
    print(arr2)
    # #
    # arr = np.load(
    #     r"H:/io_redo_share5_edit_synchronized/io_redo_share5_edit_synchronizedarrPart0.npy")
    # start = time.time()
    # arr2 = arr.T
    # print(time.time() - start)
    # print(arr2.shape)
    # print(arr2[0:25, 0])
    #
    # arr1 = np.load(
    #     r"H:/io_redo_share5_edit_synchronized/transform_arrPart0.npy")
    # print(arr1.shape)
    # print(arr1[0, 2400:2501])



import numpy as np
from  matplotlib import pyplot as plt
import scared
from numpy import loadtxt
# i_list = []此处应是数组   np.array
def snr(data,trace):
    if trace.shape[0] != data.shape[0]:
        raise ValueError("they should have same imediate value.")
    if not isinstance(trace, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    if not isinstance(data, np.ndarray):
        raise TypeError("'data' should be a numpy ndarray.")
    a = np.zeros((256,trace.shape[1]),dtype=float)
    b = np.zeros((256,trace.shape[1]),dtype=float)
    c = np.zeros((256,trace.shape[1]),dtype=float)
    a_sum = np.zeros(trace.shape[1], dtype=float)
    b_sum = np.zeros(trace.shape[1], dtype=float)
    c_sum = np.zeros(trace.shape[1], dtype=float)
    mean_std_temp = np.zeros(trace.shape[1], dtype=float)
    mean_std_temp1 = np.zeros(trace.shape[1], dtype=float)
    mean_std_temp2 = np.zeros(trace.shape[1], dtype=float)
    std_mean_temp3 = np.zeros(trace.shape[1], dtype=float)
    m = 0
    for i in range(256):
        temp1 = np.zeros(trace.shape[1], dtype=float)
        temp2 = np.zeros(trace.shape[1], dtype=float)
        temp3 = np.zeros(trace.shape[1], dtype=float)
        t_sum = np.zeros(trace.shape[1], dtype=float)
        t2_sum = np.zeros(trace.shape[1], dtype=float)

        # a[i] =
        n = 0
        for t in range(trace.shape[0]):
            if (data[t] == i):
                # 这个时候来的是trace_array[tnum]曲线
                # 需要记录下，x,x**2
                # 考虑精度，来一次除一次
                n += 1
                t_sum = temp1*(n-1) + trace[t, :]
                t2_sum = temp2*(n-1)+(trace**2)[t, :]
                # 均值
                temp1 = t_sum/n
                # 平方的均值
                temp2 = t2_sum/n
                temp3 = temp2 - temp1 ** 2
            else:
                continue
        a[i] = temp1
        b[i] = temp1**2
        m += 1
        a_sum = mean_std_temp1*(m-1)+temp1
        b_sum = mean_std_temp2*(m-1)+temp1**2
        mean_std_temp1 = a_sum/m
        mean_std_temp2 = b_sum/m
        mean_std_temp = mean_std_temp2-mean_std_temp1 ** 2
        c[i] = temp3
        c_sum = std_mean_temp3*(m-1) + temp3
        std_mean_temp3 = c_sum/m
    return mean_std_temp/std_mean_temp3
if __name__ == '__main__':

    # print(snr(data,trace))
    # plt.plot(data)
    # plt.plot(trace)
    # 太蠢了，直接传参的时候限定范围就好，函数不用改
    # data = []
    loadData = loadtxt(r"D:\traces\sendData.csv", delimiter=',',dtype="int")
    data = loadData
    loadTrace = loadtxt(r"D:\traces\trace0.csv",delimiter=',')
    # 循环读取后面的trace
    for i in range(1, 2000):
        newLoadTrace = loadtxt(
            r"D:\traces\trace{0}.csv".format(i),
            delimiter=',')
        # 拼接为矩阵
        loadTrace = np.vstack([loadTrace, newLoadTrace])

    # 最终得到的loadTrace是一个二维矩阵
    trace = scared.signal_processing.filters.butterworth(loadTrace,30e6,20e5)

    plt.plot(snr(data,trace[:,0:200]))

    plt.show()









#!/usr/bin/env python
# -*- coding: UTF-8 -*-
'''
@Project ：sca 
@File    ：suiqian.py.py
@Author  ：suyang
@Date    ：2022/5/6 16:06 
'''
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm, trange
np.seterr(divide='ignore', invalid='ignore')
def t_test(n, url_trace, trace_name):
    '''
    :instrctions:even  and  odd 间或采集
    :param n: blocks of traces
    :param N: sample numbers
    :return: result
    '''
    # arr = np.load(url_trace + r"2process_arrPart0.npy")
    arr = np.load(url_trace + trace_name + r"0.npy")
    count = 0
    N = arr.shape[1]
    old_var_even = np.zeros(N)
    old_mean_even = np.zeros(N)
    old_var_odd = np.zeros(N)
    old_mean_odd = np.zeros(N)
    oddcount = 0
    evencount = 0
    for j in trange(n):
        # arr = np.load(url_trace + r"2process_arrPart{0}.npy".format(j))
        arr = np.load(url_trace + trace_name + r"{0}.npy".format(j))
        for i in range(arr.shape[0]):
            if count % 2 == 0:
                new_mean = old_mean_even + (arr[i] - old_mean_even) / (evencount + 1)
                new_var = old_var_even + ((arr[i] - old_mean_even) * (arr[i] - new_mean) - old_var_even) / (
                        evencount + 1)
                old_mean_even = new_mean
                old_var_even = new_var
                evencount += 1
                # print(evencount)
                count = count + 1
            else:
                new_mean = old_mean_odd + (arr[i] - old_mean_odd) / (oddcount + 1)
                new_var = old_var_odd + ((arr[i] - old_mean_odd) * (arr[i] - new_mean) - old_var_odd) / (oddcount + 1)
                old_mean_odd = new_mean
                old_var_odd = new_var
                oddcount += 1
                count = count + 1
    temp1 = old_mean_even - old_mean_odd
    temp2 = (old_var_even / evencount) + (old_var_odd / oddcount)
    test_result = temp1 / np.sqrt(temp2)
    return test_result

def ttest_function(num_file_blocks, url_trace, trace_name):
    '''

    :param num_file_blocks: data is too big for our computer so we design n blocks to reduce size
    :param url_trace: data stored
    :param trace_name: As the name suggests
    :return: result
    '''
    plt.rcParams['figure.figsize'] = (12.0, 4.0)
    f, ax = plt.subplots(1, 1)
    ax.set_title('ttest_traces')
    ax.axhline(y=4.5, ls='--', c='red', linewidth=2)
    ax.axhline(y=-4.5, ls='--', c='red', linewidth=2)
    result = t_test(num_file_blocks, url_trace, trace_name)
    ax.plot(result)
    plt.show()

if __name__ == '__main__':

    # 文件路径
    # url_trace = r"H:/zhoujiayun1/"
    # 文件块数
    num_file_blocks = 100
    # 文件名字
    trace_name = "arrPart"
    # 运行函数
    # ttest_function(num_file_blocks, url_trace, trace_name)




